# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Virtual Filesystem Layer
------------------------

The ``drgn.helpers.linux.fs`` module provides helpers for working with the
Linux virtual filesystem (VFS) layer, including mounts, dentries, and inodes.
"""

import operator
import os
from typing import Iterator, Optional, Tuple, Union, overload

from drgn import (
    IntegerLike,
    Object,
    Path,
    Program,
    TypeKind,
    cast,
    container_of,
    sizeof,
)
from drgn.helpers.common.format import escape_ascii_string
from drgn.helpers.common.prog import takes_object_or_program_or_default
from drgn.helpers.linux.list import (
    hlist_empty,
    hlist_for_each_entry,
    list_for_each_entry,
)
from drgn.helpers.linux.rbtree import rbtree_inorder_for_each_entry
from drgn.helpers.linux.xarray import xa_for_each, xa_is_value

__all__ = (
    "address_space_for_each_page",
    "d_path",
    "decode_file_type",
    "dentry_path",
    "fget",
    "for_each_file",
    "for_each_mount",
    "inode_for_each_page",
    "inode_path",
    "inode_paths",
    "mount_dst",
    "mount_fstype",
    "mount_src",
    "path_lookup",
    "print_files",
    "print_mounts",
    "super_block_for_each_mount",
)


def _follow_mount(mnt: Object, dentry: Object) -> Tuple[Object, Object]:
    prog = dentry.prog_
    try:
        DCACHE_MOUNTED = prog.cache["DCACHE_MOUNTED"]
    except KeyError:
        tokens = prog["UTS_RELEASE"].string_().split(b".", 2)
        major, minor = int(tokens[0]), int(tokens[1])
        # Linux kernel commit 9748cb2dc393 ("VFS: repack DENTRY_ flags.") (in
        # v6.15) changed the value of DCACHE_MOUNTED. Unfortunately, it's a
        # macro, so we have to hardcode it based on a version check until it's
        # converted to an enum.
        if (major, minor) >= (6, 15):
            DCACHE_MOUNTED = 1 << 15
        else:
            DCACHE_MOUNTED = 1 << 16
        prog.cache["DCACHE_MOUNTED"] = DCACHE_MOUNTED

    while dentry.d_flags & DCACHE_MOUNTED:
        for mounted in list_for_each_entry(
            "struct mount", mnt.mnt_mounts.address_of_(), "mnt_child"
        ):
            if mounted.mnt_mountpoint == dentry:
                mnt = mounted
                dentry = mounted.mnt.mnt_root.read_()
                break
        else:
            break
    return mnt, dentry


def _follow_dotdot(
    mnt: Object, dentry: Object, root_mnt: Object, root_dentry: Object
) -> Tuple[Object, Object]:
    while dentry != root_dentry or mnt != root_mnt:
        d_parent = dentry.d_parent.read_()
        if dentry != d_parent:
            dentry = d_parent
            break
        mnt_parent = mnt.mnt_parent.read_()
        if mnt == mnt_parent:
            break
        dentry = mnt.mnt_mountpoint
        mnt = mnt_parent
    return _follow_mount(mnt, dentry)


@takes_object_or_program_or_default
def path_lookup(
    prog: Program,
    root: Optional[Object],
    path: Path,
    *,
    allow_negative: bool = False,
) -> Object:
    """
    Look up the given path name.

    :param root: ``struct path *`` to use as the root directory. Defaults to
        the initial root filesystem if given a :class:`~drgn.Program` or
        :ref:`omitted <default-program>`.
    :param path: Path to lookup.
    :param allow_negative: Whether to allow returning a negative dentry (i.e.,
        a dentry for a non-existent path).
    :return: ``struct path``
    :raises Exception: if the dentry is negative and ``allow_negative`` is
        ``False``, or if the path is not present in the dcache. The latter does
        not necessarily mean that the path does not exist; it may be uncached.
        On a live system, you can make the kernel cache the path by accessing
        it (e.g., with :func:`open()` or :func:`os.stat()`):

    >>> path_lookup('/usr/include/stdlib.h')
    ...
    Exception: could not find '/usr/include/stdlib.h' in dcache
    >>> open('/usr/include/stdlib.h').close()
    >>> path_lookup('/usr/include/stdlib.h')
    (struct path){
            .mnt = (struct vfsmount *)0xffff8b70413cdca0,
            .dentry = (struct dentry *)0xffff8b702ac2c480,
    }
    """
    if root is None:
        root = prog["init_task"].fs.root
    mnt = root_mnt = container_of(root.mnt.read_(), "struct mount", "mnt")
    dentry = root_dentry = root.dentry.read_()
    components = os.fsencode(path).split(b"/")
    for i, component in enumerate(components):
        if component == b"" or component == b".":
            continue
        elif component == b"..":
            mnt, dentry = _follow_dotdot(mnt, dentry, root_mnt, root_dentry)
        else:
            # Since Linux kernel commit da549bdd15c2 ("dentry: switch the lists
            # of children to hlist") (in v6.8), the children are in an hlist.
            # Before that, they're in a list with different field names.
            try:
                children = hlist_for_each_entry(
                    "struct dentry", dentry.d_children.address_of_(), "d_sib"
                )
            except AttributeError:
                children = list_for_each_entry(
                    "struct dentry", dentry.d_subdirs.address_of_(), "d_child"
                )
            for child in children:
                if child.d_name.name.string_() == component:
                    dentry = child
                    break
            else:
                failed_path = os.fsdecode(b"/".join(components[: i + 1]))
                raise Exception(f"could not find {failed_path!r} in dcache")
            mnt, dentry = _follow_mount(mnt, dentry)
    if not allow_negative and not dentry.d_inode:
        failed_path = os.fsdecode(b"/".join(components))
        raise Exception(f"{failed_path!r} dentry is negative")
    return Object(
        prog,
        "struct path",
        value={"mnt": mnt.mnt.address_of_(), "dentry": dentry},
    )


@overload
def d_path(path: Object) -> bytes:
    """
    Return the full path of a dentry given a ``struct path``.

    :param path: ``struct path`` or ``struct path *``
    """
    ...


@overload
def d_path(vfsmnt: Object, dentry: Object) -> bytes:
    """
    Return the full path of a dentry given a mount and dentry.

    :param vfsmnt: ``struct vfsmount *``
    :param dentry: ``struct dentry *``
    """
    ...


@overload
def d_path(dentry: Object) -> bytes:
    """
    Return the full path of a dentry.

    Since a mount is not provided, this arbitrarily selects a mount to determine
    the path.

    :param dentry: ``struct dentry *``
    """
    ...


def d_path(  # type: ignore  # Need positional-only arguments.
    arg1: Object, arg2: Optional[Object] = None
) -> bytes:
    if arg2 is None:
        try:
            mnt = container_of(arg1.mnt, "struct mount", "mnt")
            dentry = arg1.dentry.read_()
        except AttributeError:
            # Select an arbitrary mount from this dentry's super block. We
            # choose the first non-internal mount. Internal mounts exist for
            # kernel filesystems (e.g. debugfs) and they are mounted at "/".
            # Paths from these mounts aren't usable in userspace and they're
            # confusing. If there's no other option, we will use the first
            # internal mount we encountered.
            #
            # The MNT_INTERNAL flag is defined as a macro in the kernel source.
            # Introduced in 2.6.34 and has not been modified since.
            MNT_INTERNAL = 0x4000
            internal_mnt = None
            dentry = arg1
            for mnt in super_block_for_each_mount(dentry.d_sb):
                if mnt.mnt.mnt_flags & MNT_INTERNAL:
                    internal_mnt = internal_mnt or mnt
                    continue
                break
            else:
                if internal_mnt is not None:
                    mnt = internal_mnt
                else:
                    raise ValueError("Could not find a mount for this dentry")
    else:
        mnt = container_of(arg1, "struct mount", "mnt")
        dentry = arg2.read_()

    d_op = dentry.d_op.read_()
    if d_op and d_op.d_dname:
        return b"[" + dentry.d_inode.i_sb.s_type.name.string_() + b"]"

    components = []
    while True:
        if dentry == mnt.mnt.mnt_root:
            mnt_parent = mnt.mnt_parent.read_()
            if mnt == mnt_parent:
                break
            dentry = mnt.mnt_mountpoint.read_()
            mnt = mnt_parent
            continue
        d_parent = dentry.d_parent.read_()
        if dentry == d_parent:
            break
        components.append(dentry.d_name.name.string_())
        components.append(b"/")
        dentry = d_parent
    if components:
        return b"".join(reversed(components))
    else:
        return b"/"


def dentry_path(dentry: Object) -> bytes:
    """
    Return the path of a dentry from the root of its filesystem.

    :param dentry: ``struct dentry *``
    """
    components = []
    while True:
        d_parent = dentry.d_parent.read_()
        if dentry == d_parent:
            break
        components.append(dentry.d_name.name.string_())
        dentry = d_parent
    return b"/".join(reversed(components))


def inode_path(inode: Object) -> Optional[bytes]:
    """
    Return any path of an inode from the root of its filesystem.

    :param inode: ``struct inode *``
    :return: Path, or ``None`` if the inode has no aliases.
    """
    if hlist_empty(inode.i_dentry):
        return None
    return dentry_path(
        container_of(inode.i_dentry.first, "struct dentry", "d_u.d_alias")
    )


def inode_paths(inode: Object) -> Iterator[bytes]:
    """
    Return an iterator over all of the paths of an inode from the root of its
    filesystem.

    :param inode: ``struct inode *``
    """
    return (
        dentry_path(dentry)
        for dentry in hlist_for_each_entry(
            "struct dentry", inode.i_dentry.address_of_(), "d_u.d_alias"
        )
    )


def inode_for_each_page(inode: Object) -> Iterator[Tuple[int, Object]]:
    """
    Iterate over all cached pages and their indices in an inode.

    >>> for index, page in inode_for_each_page(inode):
    ...     print(index, hex(page))
    ...
    0 0xffffcfde4d0b6b00
    1 0xffffcfde4d0bda40
    3 0xffffcfde4d0b8b80

    :param inode: ``struct inode *``
    :return: Iterator of (index, ``struct page *`` object) tuples.
    """
    return address_space_for_each_page(inode.i_mapping.read_())


def address_space_for_each_page(mapping: Object) -> Iterator[Tuple[int, Object]]:
    """
    Iterate over all cached pages and their indices in an inode address space.

    :param mapping: ``struct address_space *``
    :return: Iterator of (index, ``struct page *`` object) tuples.
    """
    try:
        i_pages = mapping.i_pages
    except AttributeError:
        # i_pages was renamed from page_tree in Linux kernel commit
        # b93b016313b3 ("page cache: use xa_lock") (in v4.17).
        i_pages = mapping.page_tree
    page_type = mapping.prog_.type("struct page *")
    for index, entry in xa_for_each(i_pages.address_of_()):
        if not xa_is_value(entry):  # Skip shadow entries.
            yield index, cast(page_type, entry)


def mount_src(mnt: Object) -> bytes:
    """
    Get the source device name for a mount.

    :param mnt: ``struct mount *``
    """
    return mnt.mnt_devname.string_()


def mount_dst(mnt: Object) -> bytes:
    """
    Get the path of a mount point.

    :param mnt: ``struct mount *``
    """
    return d_path(mnt.mnt.address_of_(), mnt.mnt.mnt_root)


def mount_fstype(mnt: Object) -> bytes:
    """
    Get the filesystem type of a mount.

    :param mnt: ``struct mount *``
    """
    sb = mnt.mnt.mnt_sb.read_()
    fstype = sb.s_type.name.string_()
    subtype_obj = sb.s_subtype.read_()
    if subtype_obj:
        subtype = subtype_obj.string_()
        if subtype:
            fstype += b"." + subtype
    return fstype


@takes_object_or_program_or_default
def for_each_mount(
    prog: Program,
    ns: Optional[Object],
    *,
    src: Optional[Path] = None,
    dst: Optional[Path] = None,
    fstype: Optional[Union[str, bytes]] = None,
) -> Iterator[Object]:
    """
    Iterate over all of the mounts in a given namespace.

    :param ns: ``struct mnt_namespace *``. Defaults to the initial mount
        namespace if given a :class:`~drgn.Program` or :ref:`omitted
        <default-program>`.
    :param src: Only include mounts with this source device name.
    :param dst: Only include mounts with this destination path.
    :param fstype: Only include mounts with this filesystem type.
    :return: Iterator of ``struct mount *`` objects.
    """
    if ns is None:
        ns = prog["init_task"].nsproxy.mnt_ns
    if src is not None:
        src = os.fsencode(src)
    if dst is not None:
        dst = os.fsencode(dst)
    if fstype:
        fstype = os.fsencode(fstype)
    # Since Linux kernel commit 2eea9ce4310d ("mounts: keep list of mounts in
    # an rbtree") (in v6.8), the mounts in a namespace are in a red-black tree.
    # Before that, they're in a list.
    # The old case is first here because before that commit, struct mount also
    # had a different member named "mounts".
    try:
        mounts = list_for_each_entry("struct mount", ns.list.address_of_(), "mnt_list")
    except AttributeError:
        mounts = rbtree_inorder_for_each_entry(
            "struct mount", ns.mounts.address_of_(), "mnt_node"
        )
    for mnt in mounts:
        if (
            mnt.mnt.mnt_sb  # skip cursors (v5.8 - v6.8)
            and (src is None or mount_src(mnt) == src)
            and (dst is None or mount_dst(mnt) == dst)
            and (fstype is None or mount_fstype(mnt) == fstype)
        ):
            yield mnt


def super_block_for_each_mount(sb: Object) -> Iterator[Object]:
    """
    Iterate over every mount of a super block.

    :param sb: ``struct super_block *``.
    :return: Iterator of ``struct mount *`` objects.
    """
    # Since Linux kernel commit 09a1b33c080f ("preparations to taking
    # MNT_WRITE_HOLD out of ->mnt_flags") (in v6.18), mounts for a super_block
    # are on an open-coded singly-linked list. Before that, they are on a
    # list_head.
    s_mounts = sb.s_mounts
    if s_mounts.type_.unaliased_kind() == TypeKind.POINTER:
        mount = s_mounts.read_()
        while mount:
            yield mount
            mount = mount.mnt_next_for_sb.read_()
    else:
        yield from list_for_each_entry(
            "struct mount", s_mounts.address_of_(), "mnt_instance"
        )


@takes_object_or_program_or_default
def print_mounts(
    prog: Program,
    ns: Optional[Object],
    *,
    src: Optional[Path] = None,
    dst: Optional[Path] = None,
    fstype: Optional[Union[str, bytes]] = None,
) -> None:
    """
    Print the mount table of a given namespace. The arguments are the same as
    :func:`for_each_mount()`. The output format is similar to ``/proc/mounts``
    but prints the value of each ``struct mount *``.
    """
    for mnt in for_each_mount(
        prog if ns is None else ns,
        src=src,
        dst=dst,
        fstype=fstype,
    ):
        mnt_src = escape_ascii_string(mount_src(mnt), escape_backslash=True)
        mnt_dst = escape_ascii_string(mount_dst(mnt), escape_backslash=True)
        mnt_fstype = escape_ascii_string(mount_fstype(mnt), escape_backslash=True)
        print(
            f"{mnt_src} {mnt_dst} {mnt_fstype} ({mnt.type_.type_name()})0x{mnt.value_():x}"
        )


def fget(task: Object, fd: IntegerLike) -> Object:
    """
    Return the kernel file descriptor of the fd of a given task.

    :param task: ``struct task_struct *``
    :param fd: File descriptor.
    :return: ``struct file *``
    """
    return task.files.fdt.fd[fd]


def for_each_file(task: Object) -> Iterator[Tuple[int, Object]]:
    """
    Iterate over all of the files open in a given task.

    :param task: ``struct task_struct *``
    :return: Iterator of (fd, ``struct file *``) tuples.
    """
    files = task.files.read_()
    if not files:
        return
    fdt = files.fdt.read_()
    bits_per_long = 8 * sizeof(fdt.open_fds.type_.type)
    for i in range((fdt.max_fds.value_() + bits_per_long - 1) // bits_per_long):
        word = fdt.open_fds[i].value_()
        for j in range(bits_per_long):
            if word & (1 << j):
                fd = i * bits_per_long + j
                file = fdt.fd[fd].read_()
                yield fd, file


def print_files(task: Object) -> None:
    """
    Print the open files of a given task.

    :param task: ``struct task_struct *``
    """
    for fd, file in for_each_file(task):
        path = d_path(file.f_path)
        escaped_path = escape_ascii_string(path, escape_backslash=True)
        print(f"{fd} {escaped_path} ({file.type_.type_name()})0x{file.value_():x}")


# From include/uapi/linux/stat.h in the Linux kernel source code.
_S_IFMT = 0o170000
_S_IFMT_TO_STR = {
    0o140000: "SOCK",
    0o120000: "LNK",
    0o100000: "REG",
    0o060000: "BLK",
    0o040000: "DIR",
    0o020000: "CHR",
    0o010000: "FIFO",
}


def decode_file_type(mode: IntegerLike) -> str:
    """
    Convert a file mode to a human-readable file type string.

    :param mode: File mode (e.g., ``struct inode::i_mode`` or
        :attr:`os.stat_result.st_mode`).
    :return: File type as a string (e.g., "REG", "DIR", "CHR", etc.), or a raw
        octal value if unknown.
    """
    fmt = operator.index(mode) & _S_IFMT
    try:
        return _S_IFMT_TO_STR[fmt]
    except KeyError:
        return f"{fmt:06o}"



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""Functions for porting commands from :doc:`crash <crash_compatibility>`."""

import argparse
import dataclasses
import functools
import operator
import os
import re
import shlex
import shutil
import textwrap
from typing import (
    AbstractSet,
    Any,
    Callable,
    Dict,
    FrozenSet,
    Iterable,
    Iterator,
    List,
    Literal,
    NamedTuple,
    Optional,
    Sequence,
    Set,
    Tuple,
    TypeVar,
    Union,
)

from drgn import FaultError, Object, Program, ProgramFlags, Type, TypeKind, offsetof
from drgn.commands import (
    _SHELL_TOKEN_REGEX,
    DEFAULT_COMMAND_NAMESPACE,
    Command,
    CommandNamespace,
    CommandNotFoundError,
    DrgnCodeBlockContext,
    DrgnCodeBuilder,
    ParsedCommand,
    _command_name,
    _create_parser,
    _repr_black,
    argument,
    argument_group,
    command,
    custom_command,
    mutually_exclusive_group,
    unquote_shell_word,
)
from drgn.helpers.common.format import double_quote_ascii_string
from drgn.helpers.linux.cpumask import for_each_online_cpu, for_each_possible_cpu
from drgn.helpers.linux.kthread import task_is_kthread
from drgn.helpers.linux.pid import find_task, for_each_task
from drgn.helpers.linux.sched import (
    idle_task,
    task_cpu,
    task_on_cpu,
    task_state_to_char,
    thread_group_leader,
)

_PID_OR_TASK = Union[
    Tuple[Literal["pid"], int],
    Tuple[Literal["task"], int],
]


def _pid_or_task(s: str) -> _PID_OR_TASK:
    try:
        return "pid", int(s)
    except ValueError:
        return "task", int(s, 16)


_PID_OR_TASK_OR_COMMAND = Union[
    _PID_OR_TASK,
    Tuple[Literal["command"], str],
    Tuple[Literal["command_pattern"], "re.Pattern[str]"],
]


def _pid_or_task_or_command(arg: str) -> _PID_OR_TASK_OR_COMMAND:
    try:
        return "pid", int(arg, 10)
    except ValueError:
        pass

    try:
        return "task", int(arg, 16)
    except ValueError:
        pass

    if arg:
        if arg[0] == "'" and arg[-1] == "'":
            return "command_pattern", re.compile(arg[1:-1])
        if arg[0] == "\\":
            return "command", arg[1:]
    return "command", arg


_ADDR_OR_SYM = Union[
    Tuple[Literal["addr"], int],
    Tuple[Literal["sym"], str],
]


def _addr_or_sym(s: str) -> _ADDR_OR_SYM:
    try:
        return "addr", int(s, 16)
    except ValueError:
        return "sym", s


def _resolve_addr_or_sym(prog: Program, arg: _ADDR_OR_SYM) -> int:
    if arg[0] == "addr":
        return arg[1]
    else:
        return prog.symbol(arg[1]).address


def _guess_type(prog: Program, name: str, kind: str = "*") -> Type:
    if kind != "union":
        try:
            return prog.type("struct " + name)
        except LookupError:
            pass

    if kind != "struct":
        try:
            return prog.type("union " + name)
        except LookupError:
            pass

    # Try a typedef.
    type = prog.type(name)

    # Make sure it's a typedef of our desired type kind.
    unaliased_kind = type.unaliased_kind()
    if (kind != "union" and unaliased_kind == TypeKind.STRUCT) or (
        kind != "struct" and unaliased_kind == TypeKind.UNION
    ):
        return type

    if kind == "*":
        kind = "struct or union"
    raise LookupError(f"{type.type_name()} is not a {kind}")


def _guess_type_name(prog: Program, name: str, kind: str = "*") -> str:
    try:
        type = _guess_type(prog, name, kind)
    except LookupError:
        return f"{'struct' if kind == '*' else kind} {name}"
    return type.type_name()


def _object_format_options(
    prog: Program, integer_base: Optional[int]
) -> Dict[str, Any]:
    return {
        "columns": shutil.get_terminal_size().columns,
        "dereference": False,
        "integer_base": integer_base or prog.config.get("crash_radix", 10),
    }


def _format_seconds_duration(seconds: int) -> str:
    days, seconds = divmod(seconds, 86400)
    hours, seconds = divmod(seconds, 3600)
    minutes, seconds = divmod(seconds, 60)
    if days:
        return f"{days} days, {hours:02}:{minutes:02}:{seconds:02}"
    else:
        return f"{hours:02}:{minutes:02}:{seconds:02}"


def _find_pager(which: Optional[str] = None) -> Optional[str]:
    if which is None or which == "less":
        less = shutil.which("less")
        if less:
            return f"{shlex.quote(less)} -E -X"

    if which is None or which == "more":
        more = shutil.which("more")
        if more:
            return shlex.quote(more)

    return None


def _get_pager(prog: Program) -> Optional[str]:
    if not prog.config.get("crash_scroll", True):
        return None
    try:
        return prog.config["crash_pager"]
    except KeyError:
        pager = _find_pager()
        prog.config["crash_pager"] = pager
        return pager


class _CrashCommandNamespace(CommandNamespace):
    def __init__(self) -> None:
        super().__init__(
            func_name_prefix="_crash_cmd_",
            argparse_types=(
                ("pid_or_task", _pid_or_task),
                ("addr_or_sym", _addr_or_sym),
            ),
        )

    def _resolve(
        self, prog: Program, command: str, kwargs: Dict[str, Any]
    ) -> Tuple[str, str, Command[object], Optional[str]]:
        command = command.lstrip()
        if command.startswith("!"):
            return (
                "!",
                command[1:].lstrip(),
                DEFAULT_COMMAND_NAMESPACE.lookup(prog, "sh"),
                None,
            )

        if command.startswith("*"):
            name = "*"
            command_obj = self.lookup(prog, "*")
            tail = command[1:].lstrip()
        else:
            match = _SHELL_TOKEN_REGEX.match(command)
            if not match or match.lastgroup != "WORD":
                raise SyntaxError("expected command name")

            name = unquote_shell_word(match.group())
            try:
                command_obj = self.lookup(prog, name)
            except CommandNotFoundError as e:
                try:
                    # Smuggle the type into the command function.
                    kwargs["type"] = _guess_type(prog, name.partition(".")[0])
                except LookupError:
                    raise e
                else:
                    name = "*"
                    command_obj = self.lookup(prog, "*")
                    tail = command
            else:
                tail = command[match.end() :].lstrip()

        return name, tail, command_obj, None if name == "drgn" else _get_pager(prog)


CRASH_COMMAND_NAMESPACE: CommandNamespace = _CrashCommandNamespace()
"""Command namespace used for crash commands."""


crash_command = functools.partial(command, namespace=CRASH_COMMAND_NAMESPACE)
"""
Decorator to register a :doc:`crash command <crash_compatibility>`.

This is the same as :func:`~drgn.commands.command()` other than the following:

1. The command is only available as a subcommand of the :drgncommand:`crash`
   command.
2. If *name* is not given, then the name of the decorated function must begin
   with ``_crash_cmd_`` instead of ``_cmd_``.
3. The command may use the ``"pid_or_task"`` argparse type to parse a task
   context argument to a ``(type, value)`` tuple, where ``type`` is either
   ``"pid"`` or ``"task"`` and ``value`` is an :class:`int`.
"""

crash_custom_command = functools.partial(
    custom_command, namespace=CRASH_COMMAND_NAMESPACE
)
"""
Like :func:`crash_command()` but for :func:`~drgn.commands.custom_command()`.
"""

# mypy doesn't handle functools.partial(functools.partial(...), ...) very well,
# so we wrap custom_command() directly instead of raw_command().
crash_raw_command = functools.partial(
    custom_command, parse=ParsedCommand, namespace=CRASH_COMMAND_NAMESPACE
)
"""
Like :func:`crash_command()` but for :func:`~drgn.commands.raw_command()`.
"""


def _crash_get_panic_context(prog: Program) -> Object:
    if (prog.flags & (ProgramFlags.IS_LIVE | ProgramFlags.IS_LOCAL)) == (
        ProgramFlags.IS_LIVE | ProgramFlags.IS_LOCAL
    ):
        return find_task(prog, os.getpid())
    elif not (prog.flags & ProgramFlags.IS_LIVE):
        return prog.crashed_thread().object
    else:
        raise ValueError("no default context")


def _is_valid_task_struct(task: Object) -> bool:
    try:
        pid = task.pid.value_()
        if pid:
            return find_task(task.prog_, task.pid) == task
        else:
            return idle_task(task.prog_, task_cpu(task)) == task
    except FaultError:
        return False


def crash_get_context(
    prog: Program, arg: Optional[Tuple[Literal["pid", "task"], int]] = None
) -> Object:
    """
    Get the task context to use for a crash command.

    :param arg: Context parsed by the ``"pid_or_task"`` argparse type to use.
        If ``None`` or not given, use the current context.
    :return: ``struct task_struct *``
    """
    if arg is not None:
        if arg[0] == "pid":
            task = find_task(prog, arg[1])
            if not task:
                raise LookupError("no such process with PID {}".format(arg[1]))
            return task
        else:
            task = Object(prog, "struct task_struct *", arg[1])
            if not _is_valid_task_struct(task):
                raise LookupError(f"invalid task_struct: {arg[1]:#x}")
            return task

    try:
        return prog.config["crash_context"]
    except KeyError:
        pass
    prog.config["crash_context"] = task = _crash_get_panic_context(prog)
    return task


def print_task_header(task: Object) -> None:
    """Print basic information about a task in the same format as crash."""
    _print_task_header(task, cpu=task_cpu(task))


def _print_task_header(
    task: Object, *, cpu: Optional[int], prio: bool = False, end: str = "\n"
) -> None:
    if prio:
        prio_str = f"[{task.prio.value_():3}] "
    else:
        prio_str = ""
    parts = [f"{prio_str}PID: {task.pid.value_():<7}", f"TASK: {task.value_():x}"]
    if cpu is not None:
        parts.append(f"CPU: {cpu}")
    parts.append(f"COMMAND: {double_quote_ascii_string(task.comm.string_())}")
    print(*parts, sep="  ", end=end)


@dataclasses.dataclass(frozen=True)
class Cpuspec:
    """Parsed crash CPU specifier."""

    current: bool = False
    """Include the CPU of the current context."""

    all: bool = False
    """Include all possible CPUs."""

    panic: bool = False
    """Include the panic CPU."""

    explicit_cpus: FrozenSet[int] = frozenset()
    """Explicitly listed CPUs."""

    def __post_init__(self) -> None:
        if self.current + self.all + bool(self.explicit_cpus) > 1:
            raise ValueError(
                "at most one of current, all, or explicit_cpus may be given"
            )

    def cpus(self, prog: Program) -> List[int]:
        """
        Resolve the CPU specifier to a sorted list of CPU numbers, checking
        that all given CPUs were valid.
        """
        if self.current:
            return [task_cpu(crash_get_context(prog))]
        elif self.all:
            return sorted(for_each_possible_cpu(prog))
        elif self.panic:
            return [task_cpu(prog.crashed_thread().object)]
        elif self.explicit_cpus:
            possible = set(for_each_possible_cpu(prog))
            if not self.explicit_cpus.issubset(possible):
                raise ValueError(
                    f"invalid CPUs: {','.join([str(cpu) for cpu in self.explicit_cpus - possible])}"
                )
            return sorted(self.explicit_cpus)
        else:
            return []


def parse_cpuspec(spec: str) -> Cpuspec:
    """
    Parse a crash CPU specifier.

    A CPU specifier may be a comma-separated string of CPU numbers or ranges
    (e.g., '0,3-4'), 'a' or 'all' (meaning all possible CPUs), or an empty
    string (meaning the CPU of the current context).
    """
    if not spec:
        return Cpuspec(current=True)

    # Crash's parser is much more permissive: it allows extra commas (e.g.,
    # 0,,1,), extra hyphens (e.g., 0-1-2,-3--4-), and mixing "all" with CPU
    # numbers (e.g., 0-1,all). We chose to be stricter, but we can loosen it if
    # requested.
    if spec == "a" or spec == "all":
        return Cpuspec(all=True)

    cpus: Set[int] = set()
    for part in spec.split(","):
        match = re.fullmatch(r"([0-9]+)(?:-([0-9]+))?", part)
        if not match:
            raise ValueError(f"invalid cpuspec: {spec}") from None
        if match.group(2):
            cpus.update(range(int(match.group(1)), int(match.group(2)) + 1))
        else:
            cpus.add(int(match.group(1)))
    return Cpuspec(explicit_cpus=frozenset(cpus))


_TYPE_NAME_PATTERN = r"[a-zA-Z_][a-zA-Z0-9_]*"
_MEMBER_PATTERN = r"[a-zA-Z_][a-zA-Z0-9_]*(\.[a-zA-Z_][a-zA-Z0-9_]*|\[[0-9]+\])*"


# Parse a type name followed by a "." and a member name.
def _parse_type_name_and_member(arg: str) -> Tuple[str, str]:
    name, _, member = arg.partition(".")
    if not re.fullmatch(_TYPE_NAME_PATTERN, name):
        raise ValueError(f"invalid type name: {name}")
    if not re.fullmatch(_MEMBER_PATTERN, member):
        raise ValueError(f"invalid member name: {member}")
    return name, member


# Parse a type name optionally followed by a "." and one or more
# comma-separated members.
def _parse_type_name_and_members(arg: str) -> Tuple[str, List[str]]:
    name, sep, members_str = arg.partition(".")
    if not re.fullmatch(_TYPE_NAME_PATTERN, name):
        raise ValueError(f"invalid type name: {name}")
    if not sep:
        return name, []
    members = members_str.split(",")
    for member in members:
        if not re.fullmatch(_MEMBER_PATTERN, member):
            raise ValueError(f"invalid member name: {member}")
    return name, members


# Parse one or more comma-separated members.
def _parse_members(arg: str) -> List[str]:
    members = arg.split(",")
    for member in members:
        if not re.fullmatch(_MEMBER_PATTERN, member):
            raise ValueError(f"invalid member name: {member}")
    return members


# Sanitize a member name, which can contain "." and "[]" operators, to a name
# suitable for a variable.
def _sanitize_member_name(name: str) -> str:
    return re.sub(r"\.|\[([^]]+)\]", r"_\1", name)


# Parse a type offset, either as a number of bytes or a type name followed by a
# "." and a member.
def _parse_type_offset_arg(arg: str) -> Union[int, Tuple[str, str]]:
    if "." not in arg:
        try:
            return int(arg, 0)
        except ValueError:
            raise ValueError(f"invalid offset: {arg}") from None
    return _parse_type_name_and_member(arg)


# Resolve a type offset parsed with _parse_type_offset_arg() to a number of
# bytes. If match_type is given and it matches the type name, it will be used
# instead of guessing the type.
def _resolve_type_offset_arg(
    prog: Program,
    arg: Union[int, Tuple[str, str], None],
    match_type: Optional[Type] = None,
) -> int:
    if arg is None:
        return 0
    elif isinstance(arg, int):
        return arg
    else:
        name, member = arg

        if match_type is not None:
            try:
                match_name = match_type.tag
            except AttributeError:
                match_name = getattr(match_type, "name", None)
            if name == match_name:
                return offsetof(match_type, member)

        return offsetof(_guess_type(prog, name), member)


# For commands that do a symbol lookup, return whether an object lookup would
# be preferred for the sake of making --drgn output more idiomatic.
def _prefer_object_lookup(
    prog: Program, type_name: str, symbol_name: str, *, strict_type_name: bool = True
) -> bool:
    try:
        symbol_address = prog.symbol(symbol_name).address
    except LookupError:
        # If a symbol isn't found, prefer an object lookup.
        return True

    try:
        object = prog[symbol_name]
    except KeyError:
        # If an object isn't found but a symbol is, prefer a symbol lookup.
        return False

    # If both a symbol and an object are found, prefer an object lookup iff the
    # addresses are the same and the object has the desired type.
    if object.address_ != symbol_address:
        return False

    type = object.type_
    while True:
        if type.type_name() == type_name or (
            not strict_type_name and getattr(type, "tag", None) == type_name
        ):
            return True

        if type.kind != TypeKind.TYPEDEF:
            return False
        type = type.type


class CrashDrgnCodeBuilder(DrgnCodeBuilder):
    """
    Helper class for generating code for :func:`drgn_argument` for crash
    commands.
    """

    def _append_crash_panic_context(self) -> None:
        if (self._prog.flags & (ProgramFlags.IS_LIVE | ProgramFlags.IS_LOCAL)) == (
            ProgramFlags.IS_LIVE | ProgramFlags.IS_LOCAL
        ):
            self.add_import("os")
            self.add_from_import("drgn.helpers.linux.pid", "find_task")
            self.append("task = find_task(os.getpid())\n")
        else:
            self.add_from_import("drgn.helpers.linux.panic", "panic_task")
            self.append("task = panic_task()\n")

    def _append_crash_cpu_context(self, cpu: int) -> None:
        self.add_from_import("drgn.helpers.linux.sched", "cpu_curr")
        self.append(
            f"""\
cpu = {cpu}
task = cpu_curr(cpu)
"""
        )

    def _append_crash_pid_context(self, pid: int) -> None:
        self.add_from_import("drgn.helpers.linux.pid", "find_task")
        self.append(
            f"""\
pid = {pid}
task = find_task(pid)
"""
        )

    def _append_crash_task_context(self, address: int) -> None:
        self.add_from_import("drgn", "Object")
        self.append(
            f"""\
address = {hex(address)}
task = Object(prog, "struct task_struct *", address)
"""
        )

    def append_crash_context(
        self, arg: Optional[Tuple[Literal["pid", "task"], int]] = None
    ) -> None:
        """
        Append code for getting the task context in a variable named ``task``.

        :param arg: Context parsed by the ``"pid_or_task"`` argparse type to
            use. If ``None`` or not given, use the current context.
        """

        if arg is None:
            arg = self._prog.config.get("crash_context_origin")
            if arg is None:
                self._append_crash_panic_context()
                return
            elif arg[0] == "cpu":
                self._append_crash_cpu_context(arg[1])
                return

        if arg[0] == "pid":
            self._append_crash_pid_context(arg[1])
        else:
            assert arg[0] == "task"
            self._append_crash_task_context(arg[1])

    def append_task_header(
        self,
        indent: str = "",
        *,
        cpu: bool = True,
        prio: bool = False,
        variable: str = "task",
    ) -> None:
        """Append code for getting basic information about a task."""
        parts = []
        if prio:
            parts.append(f"prio = {variable}.prio\n")

        if cpu:
            self.add_from_import("drgn.helpers.linux.sched", "task_cpu")
            parts.append(f"cpu = task_cpu({variable})\n")

        parts.append(
            f"""\
pid = {variable}.pid
command = {variable}.comm
"""
        )

        self.append(textwrap.indent("".join(parts), indent))

    def append_addr_or_sym(self, arg: _ADDR_OR_SYM) -> None:
        """
        Append code for the value of an address or symbol.

        :param arg: Argument parsed by the ``"addr_or_sym"`` argparse type.
        """
        if arg[0] == "addr":
            self.append(hex(arg[1]))
        else:
            self.append(f"prog.symbol({_repr_black(arg[1])}).address")

    def begin_cpuspec_loop(self, cpuspec: Cpuspec) -> DrgnCodeBlockContext:
        """
        Begin a loop over each CPU in a CPU specifier.

        This must be paired with
        :meth:`~drgn.commands.DrgnCodeBuilder.end_block()` or used as a context
        manager.

        :param cpuspec: CPU specifier parsed by :func:`parse_cpuspec()`.
        """
        if cpuspec.current:
            self.add_from_import("drgn.helpers.linux.sched", "task_cpu")
            self.append_crash_context()
            self.append("cpu = task_cpu(task)\n")
            return self.begin_block("")
        elif cpuspec.panic:
            self.add_from_import("drgn.helpers.linux.sched", "task_cpu")
            self._append_crash_panic_context()
            self.append("cpu = task_cpu(task)\n")
            return self.begin_block("")

        if cpuspec.all:
            self.add_from_import("drgn.helpers.linux.cpumask", "for_each_possible_cpu")
            return self.begin_block("for cpu in for_each_possible_cpu():\n")
        else:
            return self.begin_block(f"for cpu in {cpuspec.cpus(self._prog)!r}:\n")

    def append_cpuspec_list(self, cpuspec: Cpuspec) -> bool:
        """
        Append code that creates a variable or list containing the CPUs in a
        CPU specifier.

        :param cpuspec: CPU specifier parsed by :func:`parse_cpuspec()`.
        :return: Whether there was potentially more than one CPU, so the code
            creates a list.
        """
        if cpuspec.current:
            self.add_from_import("drgn.helpers.linux.sched", "task_cpu")
            self.append_crash_context()
            self.append("cpu = task_cpu(task)\n")
            return False
        elif cpuspec.all:
            self.add_from_import("drgn.helpers.linux.cpumask", "for_each_possible_cpu")
            self.append("cpus = list(for_each_possible_cpu())\n")
            return True
        else:
            cpus = cpuspec.cpus(self._prog)
            if len(cpus) > 1:
                self.append(f"cpus = {cpus!r}\n")
                return True
            else:
                self.append(f"cpu = {cpus[0]}\n")
                return False


_T = TypeVar("_T")


def _in_set_condition(
    name: str, items: Iterable[_T], convert: Callable[[_T], str] = str
) -> str:
    converted = [convert(item) for item in items]
    if len(converted) == 1:
        return f"{name} == {converted[0]}"
    else:
        return f"{name} in {{{', '.join(converted)}}}"


def _not_in_set_condition(
    name: str, items: Iterable[_T], convert: Callable[[_T], str] = str
) -> str:
    converted = [convert(item) for item in items]
    if len(converted) == 1:
        return f"{name} != {converted[0]}"
    else:
        return f"{name} not in {{{', '.join(converted)}}}"


def _join_if_statement(conditions: Sequence[str], logical_op: str) -> str:
    if len(conditions) == 1:
        return f"if {conditions[0]}:\n"

    logical_op += " "
    parts = ["if (\n"]
    prefix = ""
    for condition in conditions:
        parts.append(f"    {prefix}{condition}\n")
        prefix = logical_op
    parts.append("):\n")
    return "".join(parts)


class _TaskSelector:
    def __init__(
        self,
        prog: Program,
        task_args: Sequence[Optional[_PID_OR_TASK_OR_COMMAND]] = (),
        *,
        kernel: bool = False,
        user: bool = False,
        group_leader: bool = False,
        policies: Optional[AbstractSet[int]] = None,
        on_cpu: bool = False,
        state: Optional[str] = None,
        sort: bool = False,
    ) -> None:
        self.prog = prog

        self._exact_task_args = []
        current_context = False
        self._pids = set()
        self._task_structs = set()
        self._commands = set()
        self._command_patterns = []
        for task_arg in task_args:
            if task_arg is None:
                current_context = True
            elif task_arg[0] == "pid":
                self._pids.add(task_arg[1])
            elif task_arg[0] == "task":
                self._task_structs.add(task_arg[1])
            elif task_arg[0] == "command_pattern":
                self._command_patterns.append(task_arg[1])
                continue
            else:
                self._commands.add(task_arg[1])
                continue
            self._exact_task_args.append(task_arg)

        # No commands need this, so don't bother with the added complexity.
        if current_context and (self._commands or self._command_patterns):
            raise NotImplementedError("cannot combine current context with comm filter")

        self._kernel = kernel
        self._user = user
        self._group_leader = group_leader
        self._policies = policies
        self._on_cpu = on_cpu
        self._state = state

        self._sort = sort

    def _any_filters(self) -> bool:
        return bool(
            self._commands
            or self._command_patterns
            or self._kernel
            or self._user
            or self._policies is not None
            or self._on_cpu
            or self._state is not None
        )

    def _find_exact(self) -> bool:
        return bool(
            self._exact_task_args and not self._commands and not self._command_patterns
        )

    def _filter(self, task: Object) -> Optional[Object]:
        if self._kernel and not task_is_kthread(task):
            return None
        if self._user and task_is_kthread(task):
            return None
        if self._policies is not None and task.policy.value_() not in self._policies:
            return None
        if self._on_cpu and not task_on_cpu(task):
            return None
        if self._state is not None and task_state_to_char(task) != self._state:
            return None

        if not self._exact_task_args:
            if self._group_leader and not thread_group_leader(task):
                return None
            if not self._commands and not self._command_patterns:
                return task

        if (
            # If we found the task by PID or address, then we don't need to
            # check the PID or address again.
            self._find_exact()
            # Otherwise, only read task.pid if we're filtering by PID.
            or (self._pids and task.pid.value_() in self._pids)
            or task.value_() in self._task_structs
        ):
            # If we only want group leaders and match a non-group leader by PID
            # or task_struct, then it should be replaced by its group leader.
            if self._group_leader and not thread_group_leader(task):
                return task.group_leader.read_()
            return task

        # Only read task.comm if we're filtering by comm.
        if self._commands or self._command_patterns:
            comm = task.comm.string_().decode(errors="surrogateescape")
            if comm in self._commands or any(
                pattern.search(comm) for pattern in self._command_patterns
            ):
                return task

        return None

    def tasks(self) -> Iterator[Object]:
        # If we're filtering by exact tasks and not on comm, then we can avoid
        # iterating over every task.
        if self._find_exact():
            to_sort = []
            for task_arg in self._exact_task_args:
                if task_arg == ("pid", 0):
                    for cpu in for_each_online_cpu(self.prog):
                        filtered_task = self._filter(idle_task(self.prog, cpu))
                        if filtered_task is not None:
                            if self._sort:
                                # list.sort() is stable, so these will remain
                                # in CPU order.
                                to_sort.append((0, filtered_task))
                            else:
                                yield filtered_task
                    continue

                try:
                    task = crash_get_context(self.prog, task_arg)
                except LookupError as e:
                    print(e)
                    continue

                filtered_task = self._filter(task)
                if filtered_task is not None:
                    if self._sort:
                        # Note that we sort on the original task's PID.
                        if task_arg is not None and task_arg[0] == "pid":
                            to_sort.append((task_arg[1], filtered_task))
                        else:
                            to_sort.append((task.pid.value_(), filtered_task))
                    else:
                        yield filtered_task

            if self._sort:
                to_sort.sort(key=operator.itemgetter(0))
                for _, task in to_sort:
                    yield task
            return

        unmatched_task_structs = self._task_structs.copy()
        for task in for_each_task(
            self.prog,
            # If we only want user tasks, then we don't need the idle tasks.
            # That is, unless we are filtering by task_struct, in case we need
            # them to check the validity of the given task_structs.
            idle=not self._user or bool(self._task_structs),
        ):
            unmatched_task_structs.discard(task.value_())

            filtered_task = self._filter(task)
            if filtered_task is not None:
                yield filtered_task

        for task_value in unmatched_task_structs:
            print(f"invalid task_struct: {task_value:#x}")

    def begin_task_loop(self, code: CrashDrgnCodeBuilder) -> DrgnCodeBlockContext:
        if (
            self._find_exact()
            and len(self._exact_task_args) == 1
            and self._exact_task_args[0] != ("pid", 0)
            and not self._any_filters()
        ):
            task_arg = self._exact_task_args[0]
            code.append_crash_context(task_arg)
            if task_arg is None or task_arg[0] == "pid":
                block = code.begin_block("if task:\n")
            else:
                code.append("\n")
                block = code.begin_block("")
        elif self._find_exact():
            code.append("tasks = []\n\n")
            for task_arg in self._exact_task_args:
                if task_arg == ("pid", 0):
                    code.add_from_import(
                        "drgn.helpers.linux.cpumask", "for_each_online_cpu"
                    )
                    code.add_from_import("drgn.helpers.linux.sched", "idle_task")
                    code.append(
                        """\
for cpu in for_each_online_cpu():
    tasks.append(idle_task(cpu))

"""
                    )
                else:
                    code.append_crash_context(task_arg)
                    if task_arg is None or task_arg[0] == "pid":
                        code.append("if task:\n    ")
                    code.append("tasks.append(task)\n\n")
            block = code.begin_block("for task in tasks:\n")
        else:
            code.add_from_import("drgn.helpers.linux.pid", "for_each_task")
            idle = "idle=True" if not self._user or self._task_structs else ""
            block = code.begin_block(f"for task in for_each_task({idle}):\n")

        if self._kernel:
            code.add_from_import("drgn.helpers.linux.kthread", "task_is_kthread")
            code.append("if not task_is_kthread(task):\n    continue\n")
        if self._user:
            code.add_from_import("drgn.helpers.linux.kthread", "task_is_kthread")
            code.append("if task_is_kthread(task):\n    continue\n")
        if self._policies is not None:
            condition = _not_in_set_condition(
                "task.policy.value_()", sorted(self._policies)
            )
            code.append(f"if {condition}:\n    continue\n")
        if self._on_cpu:
            code.add_from_import("drgn.helpers.linux.sched", "task_on_cpu")
            code.append("if not task_on_cpu(task):\n    continue\n")
        if self._state is not None:
            code.add_from_import("drgn.helpers.linux.sched", "task_state_to_char")
            code.append(
                f"if task_state_to_char(task) != {_repr_black(self._state)}:\n    continue\n"
            )

        block2 = None
        if self._find_exact():
            if self._group_leader:
                code.add_from_import("drgn.helpers.linux.sched", "thread_group_leader")
                code.append(
                    """\
if not thread_group_leader(task):
    task = task.group_leader.read_()
"""
                )
        elif self._pids or self._task_structs:
            if self._group_leader or self._commands or self._command_patterns:
                condition_func = _in_set_condition
                logical_op = "or"
            else:
                condition_func = _not_in_set_condition
                logical_op = "and"

            conditions = []
            if self._pids:
                conditions.append(
                    condition_func("task.pid.value_()", sorted(self._pids))
                )
            if self._task_structs:
                conditions.append(
                    condition_func("task.value_()", sorted(self._task_structs), hex)
                )
            code.append(_join_if_statement(conditions, logical_op))

            if self._group_leader:
                code.add_from_import("drgn.helpers.linux.sched", "thread_group_leader")
                code.append(
                    """\
    if not thread_group_leader(task):
        task = task.group_leader.read_()
"""
                )
            elif self._commands or self._command_patterns:
                code.append("    pass\n")

            if self._commands or self._command_patterns:
                block2 = code.begin_block("else:\n")
            else:
                if self._group_leader:
                    code.append("else:\n")
                code.append("    continue\n")
        elif self._group_leader:
            code.add_from_import("drgn.helpers.linux.sched", "thread_group_leader")
            code.append("if not thread_group_leader(task):\n    continue\n")

        if self._commands or self._command_patterns:
            code.append("comm_string = task.comm.string_().decode()\n")
            conditions = []
            if self._commands:
                conditions.append(
                    _not_in_set_condition(
                        "comm_string", sorted(self._commands), _repr_black
                    )
                )
            if self._command_patterns:
                code.add_import("re")
                for pattern in self._command_patterns:
                    conditions.append(
                        f"not re.search({_repr_black(pattern.pattern)}, comm_string)"
                    )
            code.append(_join_if_statement(conditions, "and"))
            code.append("    continue\n")

        if block2:
            block2.end()

        return block


_CrashForeachSubcommandFunc = Callable[[_TaskSelector, argparse.Namespace], Any]
_CrashForeachSubcommandFuncT = TypeVar(
    "_CrashForeachSubcommandFuncT", bound=_CrashForeachSubcommandFunc
)


class _CrashForeachSubcommand(NamedTuple):
    parser: argparse.ArgumentParser
    func: _CrashForeachSubcommandFunc


_CRASH_FOREACH_SUBCOMMANDS: Dict[str, _CrashForeachSubcommand] = {}


def _crash_foreach_subcommand(
    *,
    name: Optional[str] = None,
    arguments: Sequence[Union[argument, argument_group, mutually_exclusive_group]] = (),
) -> Callable[[_CrashForeachSubcommandFuncT], _CrashForeachSubcommandFuncT]:
    def decorator(func: _CrashForeachSubcommandFuncT) -> _CrashForeachSubcommandFuncT:
        command_name = _command_name(name, func, "_crash_foreach_")

        parser = _create_parser(
            name="foreach " + command_name,
            arguments=arguments,
            types=CRASH_COMMAND_NAMESPACE._argparse_types,
        )

        _CRASH_FOREACH_SUBCOMMANDS[command_name] = _CrashForeachSubcommand(parser, func)

        return func

    return decorator



# Copyright (c) 2025 Oracle and/or its affiliates
# SPDX-License-Identifier: LGPL-2.1-or-later

"""Implements the crash "ascii" command for drgn."""

import argparse
from typing import Any

from drgn import Program
from drgn.commands import CommandArgumentError, argument, drgn_argument
from drgn.commands._crash.common import CrashDrgnCodeBuilder, crash_command


@crash_command(
    description="translate a hexadecimal value to ASCII",
    long_description="""
    Translates 32-bit or 64-bit hexadecimal values to ASCII.
    If no argument is entered, an ASCII chart is displayed.
    """,
    arguments=(
        argument(
            "value",
            type="hexadecimal",
            nargs="*",
            help="hexadecimal value",
        ),
        drgn_argument,
    ),
)
def _crash_cmd_ascii(
    prog: Program, name: str, args: argparse.Namespace, **kwargs: Any
) -> None:
    CONTROL_NAMES = [
        "NUL",
        "SOH",
        "STX",
        "ETX",
        "EOT",
        "ENQ",
        "ACK",
        "BEL",
        "BS",
        "HT",
        "LF",
        "VT",
        "FF",
        "CR",
        "SO",
        "SI",
        "DLE",
        "DC1",
        "DC2",
        "DC3",
        "DC4",
        "NAK",
        "SYN",
        "ETB",
        "CAN",
        "EM",
        "SUB",
        "ESC",
        "FS",
        "GS",
        "RS",
        "US",
    ]

    ASCII_TABLE_TEXT = """      0    1   2   3   4   5   6   7
        +-------------------------------
    0 | NUL DLE  SP  0   @   P   '   p
    1 | SOH DC1  !   1   A   Q   a   q
    2 | STX DC2  "   2   B   R   b   r
    3 | ETX DC3  #   3   C   S   c   s
    4 | EOT DC4  $   4   D   T   d   t
    5 | ENQ NAK  %   5   E   U   e   u
    6 | ACK SYN  &   6   F   V   f   v
    7 | BEL ETB  `   7   G   W   g   w
    8 |  BS CAN  (   8   H   X   h   x
    9 |  HT  EM  )   9   I   Y   i   y
    A |  LF SUB  *   :   J   Z   j   z
    B |  VT ESC  +   ;   K   [   k   {
    C |  FF  FS  ,   <   L   \\   l   |
    D |  CR  GS  _   =   M   ]   m   }
    E |  SO  RS  .   >   N   ^   n   ~
    F |  SI  US  /   ?   O   -   o  DEL"""

    if args.drgn:
        code = CrashDrgnCodeBuilder(prog)
        code.append(
            """\
unsigned_long_size = prog.type("unsigned long").size
unsigned_long_long_size = prog.type("unsigned long long").size
for value in ["""
        )
        code.append(", ".join([hex(value) for value in args.value]))
        code.append(
            """]:
    bit_length = value.bit_length()
    if bit_length <= unsigned_long_size * 8:
        size = unsigned_long_size
    elif bit_length <= unsigned_long_long_size * 8:
        size = unsigned_long_long_size
    else:
        raise ValueError("value too large")
    bytestring = value.to_bytes(size, byteorder="little")
"""
        )

        code.print()
        return

    if not args.value:
        print(ASCII_TABLE_TEXT)
        return

    for value in args.value:
        # Choose padding width based on values of sizeof(long)/sizeof(long long)
        sizeof_long: int = prog.type("unsigned long").size  # type: ignore[assignment]
        sizeof_ull: int = prog.type("unsigned long long").size  # type: ignore[assignment]

        bit_length = value.bit_length()
        if bit_length <= sizeof_long * 8:
            prlen = sizeof_long * 2
        elif bit_length <= sizeof_ull * 8:
            prlen = sizeof_ull * 2
        else:
            raise CommandArgumentError(
                f"value too large: 0x{value:x} ({bit_length} bits vs {sizeof_ull * 8})"
            )

        bytes_ = prlen // 2
        out = []
        for i in range(bytes_):
            c = (value >> (8 * i)) & 0xFF
            if c == 0x20:
                out.append(" ")
            elif 0x21 <= c <= 0x7E:
                out.append(chr(c))
            elif c < 0x20:
                out.append(f"<{CONTROL_NAMES[c]}>")
            elif c == 0x7F:
                out.append("<DEL>")
            else:
                out.append(f"<{c:02x}>")

        # Left-pad hex output to 8 or 16 digits based on prlen, matching crash.
        print(f"{value:0{prlen}x}: {''.join(out)}")



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""Crash commands for evaluating and printing values."""

import argparse
import dataclasses
import operator
import re
import sys
from typing import Any, Callable, Dict, Optional, Tuple

from drgn import Object, Program, Type
from drgn.commands import (
    CommandArgumentError,
    ParsedCommand,
    _repr_black,
    argument,
    drgn_argument,
    parse_shell_command,
)
from drgn.commands._crash.common import (
    _MEMBER_PATTERN,
    _TYPE_NAME_PATTERN,
    CrashDrgnCodeBuilder,
    _object_format_options,
    crash_command,
    crash_custom_command,
    parse_cpuspec,
)
from drgn.helpers.linux.percpu import per_cpu

_UNITS = {
    "k": 1024,
    "K": 1024,
    "m": 1024 * 1024,
    "M": 1024 * 1024,
    "g": 1024 * 1024 * 1024,
    "G": 1024 * 1024 * 1024,
}


_UNARY_OPS: Dict[str, Callable[[Object], Object]] = {
    "-": operator.neg,
    "~": operator.inv,
}


_BINARY_OPS: Dict[str, Callable[[Object, Object], Object]] = {
    "+": operator.add,
    "-": operator.sub,
    "&": operator.and_,
    "|": operator.or_,
    "^": operator.xor,
    "*": operator.mul,
    "%": operator.mod,
    "/": operator.truediv,
    "<<": operator.lshift,
    ">>": operator.rshift,
}


_OPERAND_PATTERN = rf"""
(?P<unary{{n}}>
    (?:{'|'.join([re.escape(op) for op in _UNARY_OPS])}\s*)*
)
(?P<number{{n}}>
    (?:0[xX])?[0-9a-fA-F]+
)
(?:
    \s*
    (?P<unit{{n}}>[{''.join(_UNITS)}])
)?

"""


_EVAL_PATTERN = rf"""
    \s*
    (?P<parenthesis>\(\s*)?
    {_OPERAND_PATTERN.format(n=1)}
    (?:
        \s*
        (?P<operator>{'|'.join([re.escape(op) for op in _BINARY_OPS])})
        \s*
        {_OPERAND_PATTERN.format(n=2)}
    )?
    (?(parenthesis)\s*\))
    \s*
"""


def _eval_operand(type: Type, unary: str, number: str, unit: Optional[str]) -> Object:
    try:
        number_int = int(number, 10)
    except ValueError:
        number_int = int(number, 16)

    value = Object(type.prog, type, number_int)

    if unit is not None:
        value *= _UNITS[unit]

    for op in reversed(unary):
        if op.isspace():
            continue
        value = _UNARY_OPS[op](value)

    return value


def _append_operand(
    code: CrashDrgnCodeBuilder,
    type_name: str,
    unary: str,
    number: str,
    unit: Optional[str],
) -> None:
    for op in unary:
        if not op.isspace():
            code.append(op)

    if unit is not None:
        code.append("(")

    if re.match(r"[0-9]*[a-fA-F]", number):
        number = "0x" + number

    code.append(f'Object(prog, "{type_name}", {number})')

    if unit is not None:
        code.append(f" * {_UNITS[unit]})")


def _parse_eval(args: str) -> ParsedCommand[str]:
    # This command allows redirection and pipelines. To avoid ambiguity with
    # the <<, >>, and | operators, those operators must be parenthesized.
    # Split the arguments from any redirections/pipelines. We count nested
    # parentheses even though the expression syntax currently doesn't support
    # nesting.
    parens = 0
    for i, c in enumerate(args):
        if c == "(":
            parens += 1
        elif parens:
            if c == ")":
                parens -= 1
        elif c in "<>|":
            break
    else:
        i = len(args)

    parsed = parse_shell_command(args[i:])
    if parsed.args:
        # Don't allow extra arguments to be mixed in with redirections.
        raise SyntaxError("eval does not support arguments after redirections")
    return dataclasses.replace(parsed, args=args[:i])  # type: ignore[return-value]


@crash_custom_command(
    description="evaluate an expression",
    usage="**eval** [**-b**] [**-l**] [**--drgn**] (*expression*)",
    long_description="Evaluate an expression and print the result in various formats.",
    arguments=(
        argument(
            "-b",
            action="store_true",
            help="show which bits are set in the result, where 0 is the least significant bit",
        ),
        argument(
            "-l",
            action="store_true",
            help="use 64-bit values even on 32-bit architectures",
        ),
        argument(
            "expression",
            help=r"""
        expression to evaluate.

        Expressions are of the form ``value`` or ``value operator value``,
        where value is a number optionally prefixed by zero or more unary
        operators and optionally suffixed by a unit.

        Numbers may be decimal or hexadecimal. The "0x" prefix is optional for
        hexadecimal numbers.

        The following binary operators are supported:

        ``+  -  &  |  ^  *  %  /  <<  >>``

        The following unary operators are supported:

        ``- ~``

        Units may be "K" or "k" to multiply by 1024, "M" or "m" to multiply by
        1024*1024, and "G" or "g" to multiply by 1024*1024*1024.

        Expressions may be enclosed in parentheses. Parentheses are only
        required when using the ``|``, ``<<``, or ``>>`` operators.
        """,
        ),
        drgn_argument,
    ),
    parse=_parse_eval,
)
def _crash_cmd_eval(prog: Program, name: str, args: str, **kwargs: Any) -> None:
    # Handle (and remove) the --drgn flag after the expression.
    expr, n = re.subn(r"((^|\s+)--drgn)+\s*$", "", args)
    drgn_arg = n > 0

    # Handle (and remove) command line flags (-b, -l, and --drgn) before
    # the expression.
    show_bits_set = False
    long_long = False

    i = 0
    pattern = re.compile(
        r"""
        \s*
        (?:
            # -l and --drgn are unambiguous.
            (-[bl]*l[bl]*|--drgn)(?:\s|$)
            |
            # -b could be a flag or an expression (equivalent to -0xb).
            # Crash treats it as a flag unless it is the last argument, so
            # we do the same.
            (-b)\s+\S
        )
        """,
        flags=re.VERBOSE,
    )
    while match := pattern.match(expr, pos=i):
        flags = match.group(match.lastindex)  # type: ignore[arg-type]  # lastindex cannot be None
        if flags == "--drgn":
            drgn_arg = True
        else:
            show_bits_set |= "b" in flags
            long_long |= "l" in flags
        i = match.end(match.lastindex)  # type: ignore[arg-type]  # lastindex cannot be None
    expr = expr[i:]

    match = re.fullmatch(_EVAL_PATTERN, expr, flags=re.VERBOSE)
    if not match:
        raise CommandArgumentError("bad expression")

    type_name = "unsigned long long" if long_long else "unsigned long"
    op = match.group("operator")

    if drgn_arg:
        code = CrashDrgnCodeBuilder(prog)
        code.add_from_import("drgn", "Object", "cast")
        code.add_from_import("drgn.helpers.common.format", "number_in_binary_units")
        if op is None:
            code.append("value = ")
            _append_operand(
                code,
                type_name,
                match.group("unary1"),
                match.group("number1"),
                match.group("unit1"),
            )
            code.append("\n")
        else:
            code.append("lhs = ")
            _append_operand(
                code,
                type_name,
                match.group("unary1"),
                match.group("number1"),
                match.group("unit1"),
            )
            code.append("\nrhs = ")
            _append_operand(
                code,
                type_name,
                match.group("unary2"),
                match.group("number2"),
                match.group("unit2"),
            )
            code.append(f"\nvalue = lhs {op} rhs\n")
        code.append(
            f"""
signed = cast("{type_name.replace("unsigned ", "")}", value)
in_units = number_in_binary_units(value)
in_hex = hex(value)
in_octal = oct(value)
in_binary = bin(value)
"""
        )
        if show_bits_set:
            code.add_from_import("drgn", "sizeof")
            code.append(
                """\
bits_set = [
i for i in range(sizeof(value) * 8) if value & (1 << i)
]
"""
            )
        code.print()
        return

    type = prog.type(type_name)
    bit_size = type.size * 8  # type: ignore[operator]  # type.size cannot be None.

    obj = _eval_operand(
        type, match.group("unary1"), match.group("number1"), match.group("unit1")
    )
    if op is not None:
        rhs = _eval_operand(
            type,
            match.group("unary2"),
            match.group("number2"),
            match.group("unit2"),
        )
        obj = _BINARY_OPS[op](obj, rhs)

    value = obj.value_()

    with_units = ""
    if value % (1024 * 1024 * 1024) == 0:
        with_units = f"  ({value // (1024 * 1024 * 1024)}GB)"
    elif value % (1024 * 1024) == 0:
        with_units = f"  ({value // (1024 * 1024)}MB)"
    elif value % 1024 == 0:
        with_units = f"  ({value // 1024}KB)"

    signed = ""
    if value & (1 << (bit_size - 1)):
        signed = f"  ({value - (1 << bit_size)})"

    bits_set = ""
    if show_bits_set:
        bits_set = "\n   bits set:" + "".join(
            [f" {i}" for i in range(bit_size - 1, -1, -1) if value & (1 << i)]
        )

    sys.stdout.write(
        f"""\
hexadecimal: {value:x}{with_units}
    decimal: {value}{signed}
      octal: {value:o}
     binary: {value:0{bit_size}b}{bits_set}
"""
    )


def _parse_name_and_optional_member(s: str) -> Tuple[str, str]:
    name, sep, member = s.partition(".")
    if not re.fullmatch(_TYPE_NAME_PATTERN, name) or not re.fullmatch(
        _MEMBER_PATTERN, member
    ):
        return s, ""
    return name, member


@crash_command(
    description="print the value of an object",
    arguments=(
        argument(
            "object",
            metavar="object[:cpuspec]",
            help="object to print. "
            "This may include member accesses and array subscripts. "
            "It does not support arbitrary expressions yet. "
            "For per-cpu variables, this may also contain a colon (':') "
            "followed by a specification of which CPUs to print, "
            "which may be a comma-separated string of CPU numbers or ranges "
            "(e.g., '0,3-4'), "
            "'a' or 'all' (meaning all possible CPUs), "
            "or an empty string (meaning the CPU of the current context)",
        ),
        argument(
            "-x",
            dest="integer_base",
            action="store_const",
            const=16,
            help="output integers in hexadecimal format regardless of the default",
        ),
        argument(
            "-d",
            dest="integer_base",
            action="store_const",
            const=10,
            help="output integers in decimal format regardless of the default",
        ),
        drgn_argument,
    ),
)
def _crash_cmd_p(
    prog: Program, name: str, args: argparse.Namespace, **kwargs: Any
) -> None:
    expr, sep, cpuspec_str = args.object.partition(":")
    name, member = _parse_name_and_optional_member(expr)
    cpuspec = parse_cpuspec(cpuspec_str) if sep else None

    if args.drgn:
        if member:
            member = "." + member
        if cpuspec is None:
            print(f"object = prog[{_repr_black(name)}]{member}")
        else:
            code = CrashDrgnCodeBuilder(prog)
            code.append(f"pcpu_object = prog[{_repr_black(name)}]{member}\n")
            code.add_from_import("drgn.helpers.linux.percpu", "per_cpu")
            with code.begin_cpuspec_loop(cpuspec):
                code.append("object = per_cpu(pcpu_object, cpu)\n")
            code.print()
        return

    format_options = _object_format_options(prog, args.integer_base)
    obj = prog[name]
    if member:
        obj = obj.subobject_(member)
    if cpuspec is None:
        print(f"{expr} = {obj.format_(**format_options)}")
    else:
        for cpu in cpuspec.cpus(prog):
            print(
                f"per_cpu({expr}, {cpu}) = {per_cpu(obj, cpu).format_(**format_options)}"
            )



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""Crash commands for inspecting structures and unions."""

import argparse
import functools
import sys
from typing import Any, Iterable, List, Optional, Tuple, Union

from drgn import Object, Program, offsetof, sizeof
from drgn.commands import CommandError, _repr_black, argument, drgn_argument
from drgn.commands._crash.common import (
    _ADDR_OR_SYM,
    _PID_OR_TASK,
    Cpuspec,
    CrashDrgnCodeBuilder,
    _addr_or_sym,
    _crash_foreach_subcommand,
    _guess_type,
    _guess_type_name,
    _object_format_options,
    _parse_members,
    _parse_type_name_and_members,
    _parse_type_offset_arg,
    _pid_or_task,
    _prefer_object_lookup,
    _resolve_addr_or_sym,
    _sanitize_member_name,
    _TaskSelector,
    crash_command,
    parse_cpuspec,
    print_task_header,
)
from drgn.helpers.linux.percpu import per_cpu_ptr
from drgn.helpers.linux.sched import task_thread_info


# Workaround for https://github.com/python/cpython/issues/80259 from
# https://github.com/python/cpython/issues/80259#issuecomment-1093816101.
@functools.wraps(int)
def _int_or_suppress(s: str) -> Union[int, str]:
    return s if s is argparse.SUPPRESS else int(s)


def _struct_drgn_option(
    prog: Program,
    args: argparse.Namespace,
    *,
    name: str,
    members: List[str],
    type_name: str,
    offset_arg: Union[None, int, Tuple[str, str]],
    address_or_symbol: Optional[_ADDR_OR_SYM],
    cpuspec: Optional[Cpuspec],
) -> None:
    if address_or_symbol is None:
        if members:
            sys.stdout.write(
                f"""\
from drgn import offsetof
from drgn.helpers.common.type import typeof_member


type = prog.type("{type_name}")
"""
            )
            for member in members:
                sys.stdout.write(
                    f"""
{_sanitize_member_name(member)}_type = typeof_member(type, "{member}")
{_sanitize_member_name(member)}_offset = offsetof(type, "{member}")
"""
                )
        else:
            sys.stdout.write(
                f"""\
from drgn import sizeof


type = prog.type("{type_name}")
size = sizeof(type)
"""
            )
        return

    code = CrashDrgnCodeBuilder(prog)

    object_or_pointer = "object" if args.count == 1 else "pointer"
    per_cpu_helper = "per_cpu" if object_or_pointer == "object" else "per_cpu_ptr"
    pcpu_prefix = "" if cpuspec is None else "pcpu_"

    if isinstance(offset_arg, tuple):

        def initial_object(address: str) -> str:
            return f'offset_pointer = Object(prog, "void *", {address})'

    elif object_or_pointer == "object":

        def initial_object(address: str) -> str:
            return (
                f'{pcpu_prefix}object = Object(prog, "{type_name}", address={address})'
            )

    else:

        def initial_object(address: str) -> str:
            return f'{pcpu_prefix}pointer = Object(prog, "{type_name} *", {address})'

    if isinstance(offset_arg, int):
        subtract_offset = f" - {offset_arg}"
    else:
        subtract_offset = ""

    if address_or_symbol[0] == "addr":
        code.add_from_import("drgn", "Object")
        code.append(f"{initial_object(hex(address_or_symbol[1]) + subtract_offset)}\n")
    else:
        symbol_name = address_or_symbol[1]
        # This command technically always does a symbol lookup, but prefer
        # recommending a prog[] lookup if it is equivalent.
        if offset_arg is None and _prefer_object_lookup(prog, type_name, symbol_name):
            code.append(
                f"{pcpu_prefix}{object_or_pointer} = prog[{_repr_black(symbol_name)}]"
            )
            if object_or_pointer == "pointer":
                code.append(".address_of_()")
            code.append("\n")
        else:
            code.add_from_import("drgn", "Object")
            code.append(
                f"""\
address = prog.symbol({_repr_black(symbol_name)}).address{subtract_offset}
{initial_object("address")}
"""
            )

    if isinstance(offset_arg, tuple):
        offset_name, offset_member = offset_arg
        after = "[0]" if object_or_pointer == "object" else ""
        code.add_from_import("drgn", "container_of")
        if name == offset_name:
            offset_type_name = type_name
        else:
            code.add_from_import("drgn", "reinterpret")
            offset_type_name = _guess_type_name(prog, offset_name)
            if object_or_pointer == "object":
                after += f'\n{pcpu_prefix}object = reinterpret("{type_name}", {pcpu_prefix}object)'
            else:
                after += f'\n{pcpu_prefix}pointer = cast("{type_name} *", {pcpu_prefix}pointer)'
        code.append(
            f'{pcpu_prefix}{object_or_pointer} = container_of(offset_pointer, "{offset_type_name}", "{offset_member}"){after}\n'
        )

    if cpuspec is not None:
        code.begin_cpuspec_loop(cpuspec)
        code.add_from_import("drgn.helpers.linux.percpu", per_cpu_helper)
        code.append(
            f"{object_or_pointer} = {per_cpu_helper}({pcpu_prefix}{object_or_pointer}, cpu)\n"
        )

    if args.count != 1:
        if args.count >= 0:
            slice_str = f":{args.count}"
        else:
            slice_str = f"{args.count + 1}:1"
        code.begin_block(f"for object in pointer[{slice_str}]:\n")

    code.append(
        "".join(
            [
                f"{_sanitize_member_name(member)} = object.{member}\n"
                for member in members
            ]
        )
    )

    if args.count != 1:
        if not members:
            code.append("...\n")
        code.end_block()

    if cpuspec is not None:
        code.end_block()

    code.print()
    return


@crash_command(
    description="structure contents",
    arguments=(
        argument(
            "name",
            metavar="struct_name[.member[,member]]",
            help="name of structure type; one or more comma-separated members "
            "(each of which can be nested and include array subscripts) "
            "may also be given to limit the output to those members",
        ),
        argument(
            "address_or_symbol",
            metavar="address_or_symbol[:cpuspec]",
            nargs="?",
            help="hexadecimal address or symbol name of structure. "
            "If not given, the type and its size are printed instead. "
            "For per-cpu variables, this may also contain a colon (':') "
            "followed by a specification of which CPUs to print, "
            "which may be a comma-separated string of CPU numbers or ranges "
            "(e.g., '0,3-4'), "
            "'a' or 'all' (meaning all possible CPUs), "
            "or an empty string (meaning the CPU of the current context)",
        ),
        argument(
            "-l",
            dest="offset",
            help="offset from the beginning of the desired structure to the "
            "given address or symbol, either as a number of bytes or a "
            "struct_name.member",
        ),
        argument(
            "-x",
            dest="integer_base",
            action="store_const",
            const=16,
            help="output integers in hexadecimal format regardless of the default",
        ),
        argument(
            "-d",
            dest="integer_base",
            action="store_const",
            const=10,
            help="output integers in decimal format regardless of the default",
        ),
        argument("-c", dest="count", type=int, help="number of consecutive structures"),
        argument(
            "count",
            type=_int_or_suppress,
            nargs="?",
            default=argparse.SUPPRESS,
            help="number of consecutive structures",
        ),
        drgn_argument,
    ),
)
def _crash_cmd_struct(
    prog: Program, kind: str, args: argparse.Namespace, **kwargs: Any
) -> None:
    name, members = _parse_type_name_and_members(args.name)
    offset_arg = None if args.offset is None else _parse_type_offset_arg(args.offset)

    if args.address_or_symbol is None:
        if args.count is not None:
            raise CommandError("-c/count requires address or symbol")
        if args.offset is not None:
            raise CommandError("-l requires address or symbol")
        address_or_symbol = None
        cpuspec = None
    else:
        if args.count is None:
            args.count = 1
        address_or_symbol_str, sep, cpuspec_str = args.address_or_symbol.partition(":")
        address_or_symbol = _addr_or_sym(address_or_symbol_str)
        cpuspec = parse_cpuspec(cpuspec_str) if sep else None

    # We look up the type even for --drgn so that we can get the correct type
    # name (e.g., "atomic_t", not "struct atomic_t") if it exists. If it
    # doesn't, then it's only a hard error without --drgn.
    try:
        # If this was run via an implicit type command, then the type should
        # already be smuggled in here.
        try:
            type = kwargs["type"]
        except KeyError:
            type = _guess_type(prog, name, kind)
    except LookupError:
        if not args.drgn:
            raise
        type_name = f"{'struct' if kind == '*' else kind} {name}"
    else:
        type_name = type.type_name()

    if args.drgn:
        _struct_drgn_option(
            prog,
            args,
            name=name,
            members=members,
            type_name=type_name,
            offset_arg=offset_arg,
            address_or_symbol=address_or_symbol,
            cpuspec=cpuspec,
        )
        return

    if address_or_symbol is None:
        if members:
            # Crash doesn't support multiple members here, but we easily can.
            # Crash also doesn't support nested members and silently fails for
            # array subscripts. We can support these by sanitizing the member
            # name and adding a comment with the original name.
            for i, member in enumerate(members):
                obj = Object(prog, type, address=0).subobject_(member)
                sanitized = _sanitize_member_name(member)
                if sanitized == member:
                    sanitized_comment = ""
                else:
                    sanitized_comment = f" /* {member} */"
                print(
                    f"[{obj.address_}] {obj.type_.variable_declaration(sanitized)};{sanitized_comment}"
                )
        else:
            sys.stdout.write(f"{type}\nSIZE: {sizeof(type)}\n")
        return

    if offset_arg is None:
        offset = 0
    elif isinstance(offset_arg, int):
        offset = offset_arg
    else:
        offset_name, offset_member = offset_arg
        if offset_name == name:
            offset_type = type
        else:
            offset_type = _guess_type(prog, offset_name)
        offset = offsetof(offset_type, offset_member)

    address = _resolve_addr_or_sym(prog, address_or_symbol) - offset

    if args.count >= 0:
        sl = slice(0, args.count)
    else:
        sl = slice(args.count + 1, 1)

    ptr = Object(prog, prog.pointer_type(type), address)

    if cpuspec is None:

        def arrays() -> Iterable[Object]:
            return (ptr[sl],)

    else:

        def arrays() -> Iterable[Object]:
            for cpu in cpuspec.cpus(prog):
                pcpu_ptr = per_cpu_ptr(ptr, cpu)
                print(f"[{cpu}]: {pcpu_ptr.value_():x}")
                yield pcpu_ptr[sl]

    format_options = _object_format_options(prog, args.integer_base)
    for arr in arrays():
        for i, obj in enumerate(arr):
            if i != 0:
                print()

            if members:
                for member in members:
                    print(
                        f"{member} = {obj.subobject_(member).format_(**format_options)}"
                    )
            else:
                print(obj.format_(**format_options))


@crash_command(
    description="union contents",
    arguments=(
        argument(
            "name",
            metavar="union_name[.member[,member]]",
            help="name of union type; one or more comma-separated members "
            "(each of which can be nested and include array subscripts) "
            "may also be given to limit the output to those members",
        ),
        argument(
            "address_or_symbol",
            metavar="address_or_symbol[:cpuspec]",
            nargs="?",
            help="hexadecimal address or symbol name of union. "
            "If not given, the type and its size are printed instead. "
            "For per-cpu variables, this may also contain a colon (':') "
            "followed by a specification of which CPUs to print, "
            "which may be a comma-separated string of CPU numbers or ranges "
            "(e.g., '0,3-4'), "
            "'a' or 'all' (meaning all possible CPUs), "
            "or an empty string (meaning the CPU of the current context)",
        ),
        argument(
            "-l",
            dest="offset",
            help="offset from the beginning of the desired union to the "
            "given address or symbol, either as a number of bytes or a "
            "union_name.member",
        ),
        argument(
            "-x",
            dest="integer_base",
            action="store_const",
            const=16,
            help="output integers in hexadecimal format regardless of the default",
        ),
        argument(
            "-d",
            dest="integer_base",
            action="store_const",
            const=10,
            help="output integers in decimal format regardless of the default",
        ),
        argument("-c", dest="count", type=int, help="number of consecutive unions"),
        argument(
            "count",
            type=_int_or_suppress,
            nargs="?",
            default=argparse.SUPPRESS,
            help="number of consecutive unions",
        ),
        drgn_argument,
    ),
)
def _crash_cmd_union(
    prog: Program, name: str, args: argparse.Namespace, **kwargs: Any
) -> None:
    return _crash_cmd_struct(prog, name, args, **kwargs)


@crash_command(
    name="*",
    description="shortcut for struct or union",
    usage=r"\* [*struct or union command arguments*]",
    long_description="""
    This is a shortcut that allows typing, e.g., ``*list_head`` instead of
    ``struct list_head``. Note that if the type name is not also the name of a
    command, then the ``*`` can also be omitted, e.g., ``list_head``.
    """,
    arguments=(
        argument(
            "name", metavar="struct_name[.member[,member]]", help=argparse.SUPPRESS
        ),
        argument(
            "address_or_symbol",
            metavar="address_or_symbol[:cpuspec]",
            nargs="?",
            help=argparse.SUPPRESS,
        ),
        argument("-l", dest="offset", help=argparse.SUPPRESS),
        argument("-c", dest="count", type=int, help=argparse.SUPPRESS),
        argument(
            "-x",
            dest="integer_base",
            action="store_const",
            const=16,
            help=argparse.SUPPRESS,
        ),
        argument(
            "-d",
            dest="integer_base",
            action="store_const",
            const=10,
            help=argparse.SUPPRESS,
        ),
        argument(
            "count",
            type=_int_or_suppress,
            nargs="?",
            default=argparse.SUPPRESS,
            help=argparse.SUPPRESS,
        ),
        argument("--drgn", action="store_true", help=argparse.SUPPRESS),
    ),
)
def _crash_cmd_asterisk(
    prog: Program, name: str, args: argparse.Namespace, **kwargs: Any
) -> None:
    return _crash_cmd_struct(prog, name, args, **kwargs)


@_crash_foreach_subcommand(
    arguments=(
        argument(
            "-R",
            dest="members",
            metavar="member[,member]",
            action="append",
        ),
        argument(
            "-x",
            dest="integer_base",
            action="store_const",
            const=16,
        ),
        argument(
            "-d",
            dest="integer_base",
            action="store_const",
            const=10,
        ),
        drgn_argument,
    ),
)
def _crash_foreach_task(task_selector: _TaskSelector, args: argparse.Namespace) -> None:
    prog = task_selector.prog

    members = []
    if args.members:
        for arg in args.members:
            members.extend(_parse_members(arg))

    if args.drgn:
        code = CrashDrgnCodeBuilder(prog)
        with task_selector.begin_task_loop(code):
            code.append_task_header()
            if members:
                for member in members:
                    code.append(f"{_sanitize_member_name(member)} = task.{member}\n")
            else:
                code.add_from_import("drgn.helpers.linux.sched", "task_thread_info")
                code.append("thread_info = task_thread_info(task)\n")
        return code.print()

    format_options = _object_format_options(prog, args.integer_base)

    first = True
    for task in task_selector.tasks():
        if first:
            first = False
        else:
            print()

        print_task_header(task)
        if members:
            for member in members:
                print(
                    f"  {member} = {task[0].subobject_(member).format_(**format_options)}"
                )
        else:
            print(
                task[0].format_(**format_options),
                task_thread_info(task)[0].format_(**format_options),
                sep="\n\n",
            )


@crash_command(
    description="task_struct and thread_info contents",
    arguments=(
        argument(
            "tasks",
            metavar="pid|task",
            nargs="*",
            help="""
            display this task, given as either a decimal process ID or a
            hexadecimal ``task_struct`` address. May be given multiple times.
            Defaults to the current context
            """,
        ),
        argument(
            "-R",
            dest="members",
            metavar="member[,member]",
            action="append",
            help="""
            display only these ``task_struct`` members, given as a
            comma-separated list. Each member can be nested and include array
            subscripts. The **-R** is optional
            """,
        ),
        argument(
            "-x",
            dest="integer_base",
            action="store_const",
            const=16,
            help="output integers in hexadecimal format regardless of the default",
        ),
        argument(
            "-d",
            dest="integer_base",
            action="store_const",
            const=10,
            help="output integers in decimal format regardless of the default",
        ),
        drgn_argument,
    ),
)
def _crash_cmd_task(
    prog: Program, name: str, args: argparse.Namespace, **kwargs: Any
) -> None:
    task_args: List[Optional[_PID_OR_TASK]] = []
    for arg in args.tasks:
        try:
            task_args.append(_pid_or_task(arg))
        except ValueError:
            if args.members is None:
                args.members = [arg]
            else:
                args.members.append(arg)

    if not task_args:
        task_args.append(None)
    return _crash_foreach_task(_TaskSelector(prog, task_args), args)



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
PCI
---

The ``drgn.helpers.linux.pci`` module provides helpers for working with PCI
devices and buses.
"""

import enum
import os
from typing import Iterator, Union

from drgn import Object, Program, container_of
from drgn.helpers.common.prog import takes_program_or_default
from drgn.helpers.linux.device import bus_for_each_dev, dev_name
from drgn.helpers.linux.list import list_for_each_entry

__all__ = (
    "for_each_pci_dev",
    "for_each_pci_root_bus",
    "pci_bus_for_each_child",
    "pci_bus_for_each_dev",
    "pci_bus_name",
    "pci_is_bridge",
    "pci_name",
    "pci_pcie_type",
)


@takes_program_or_default
def for_each_pci_dev(prog: Program) -> Iterator[Object]:
    """
    Iterate over all PCI devices.

    :return: Iterator of ``struct pci_dev *`` objects.
    """
    pci_dev_type = prog.type("struct pci_dev")
    for dev in bus_for_each_dev(prog["pci_bus_type"].address_of_()):
        yield container_of(dev, pci_dev_type, "dev")


@takes_program_or_default
def for_each_pci_root_bus(prog: Program) -> Iterator[Object]:
    """
    Iterate over all PCI root buses.

    :return: Iterator of ``struct pci_bus *`` objects.
    """
    return list_for_each_entry(
        "struct pci_bus", prog["pci_root_buses"].address_of_(), "node"
    )


def pci_bus_for_each_child(bus: Object) -> Iterator[Object]:
    """
    Iterate over every child bus of a given PCI bus.

    :param bus: ``struct pci_bus *``
    :return: Iterator of ``struct pci_bus *`` objects.
    """
    return list_for_each_entry("struct pci_bus", bus.children.address_of_(), "node")


def pci_bus_for_each_dev(bus: Object) -> Iterator[Object]:
    """
    Iterate over every device on a given PCI bus.

    :param bus: ``struct pci_bus *``
    :return: Iterator of ``struct pci_dev *`` objects.
    """
    return list_for_each_entry("struct pci_dev", bus.devices.address_of_(), "bus_list")


def pci_name(dev: Object) -> str:
    """
    Get the name (Domain:Bus:Device.Function) of a PCI device.

    :param dev: ``struct pci_dev *``
    """
    return os.fsdecode(dev_name(dev.dev))


def pci_bus_name(bus: Object) -> str:
    """
    Get the name (Domain:Bus) of a PCI bus.

    :param bus: ``struct pci_bus *``
    """
    return os.fsdecode(dev_name(bus.dev))


_PCI_HEADER_TYPE_NORMAL = 0
_PCI_HEADER_TYPE_BRIDGE = 1
_PCI_HEADER_TYPE_CARDBUS = 2


def pci_is_bridge(dev: Object) -> bool:
    """
    Return whether a PCI device is a bridge.

    :param dev: ``struct pci_dev *``
    """
    return dev.hdr_type.value_() in (_PCI_HEADER_TYPE_BRIDGE, _PCI_HEADER_TYPE_CARDBUS)


def pci_pcie_type(dev: Object) -> Union["PCI_EXP_TYPE", int]:
    """
    Return the PCI Express Device/Port Type of a PCI device.

    :param dev: ``struct pci_dev *``
    :return: :class:`PCI_EXP_TYPE` constant, or an ``int`` if the value is not
        recognized.
    """
    type = (dev.pcie_flags_reg.value_() & 0x00F0) >> 4
    try:
        return PCI_EXP_TYPE(type)
    except ValueError:
        return type


class PCI_EXP_TYPE(enum.IntEnum):
    """PCI Express Device/Port Type."""

    ENDPOINT = 0x0
    """Express Endpoint"""
    LEG_END = 0x1
    """Legacy Endpoint"""
    ROOT_PORT = 0x4
    """Root Port"""
    UPSTREAM = 0x5
    """Upstream Port"""
    DOWNSTREAM = 0x6
    """Downstream Port"""
    PCI_BRIDGE = 0x7
    """PCIe to PCI/PCI-X Bridge"""
    PCIE_BRIDGE = 0x8
    """PCI/PCI-X to PCIe Bridge"""
    RC_END = 0x9
    """Root Complex Integrated Endpoint"""
    RC_EC = 0xA
    """Root Complex Event Collector"""



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Slab Allocator
--------------

The ``drgn.helpers.linux.slab`` module provides helpers for working with the
Linux slab allocator.

.. warning::

    Beware of slab merging when using these helpers. See
    :func:`slab_cache_is_merged()
    <drgn.helpers.linux.slab.slab_cache_is_merged>`.
"""

import functools
import operator
import os
from typing import Callable, Dict, Iterator, NamedTuple, Optional, Set, Tuple, Union

from drgn import (
    NULL,
    FaultError,
    IntegerLike,
    Object,
    ObjectNotFoundError,
    Program,
    ProgramFlags,
    Type,
    cast,
    container_of,
    sizeof,
)
from drgn.helpers import ValidationError
from drgn.helpers.common.format import escape_ascii_string
from drgn.helpers.common.prog import takes_program_or_default
from drgn.helpers.linux.cpumask import for_each_online_cpu
from drgn.helpers.linux.list import list_for_each_entry, validate_list_for_each_entry
from drgn.helpers.linux.mm import (
    PageSlab,
    _get_PageSlab_impl,
    compound_head,
    for_each_page,
    in_direct_map,
    page_to_virt,
    virt_to_page,
)
from drgn.helpers.linux.nodemask import for_each_online_node, nr_node_ids
from drgn.helpers.linux.percpu import per_cpu_ptr
from drgn.helpers.linux.rbtree import rbtree_inorder_for_each_entry
from drgn.helpers.linux.vmstat import global_node_page_state

__all__ = (
    "find_containing_slab_cache",
    "find_slab_cache",
    "for_each_slab_cache",
    "get_slab_cache_aliases",
    "print_slab_caches",
    "slab_cache_for_each_allocated_object",
    "slab_cache_is_merged",
    "slab_cache_objects_per_slab",
    "slab_cache_order",
    "slab_cache_pages_per_slab",
    "slab_cache_usage",
    "slab_object_info",
    "slab_total_usage",
)


# Get the type containing slab information.
#
# Linux kernel commit d122019bf061cccc4583eb9ad40bf58c2fe517be ("mm: Split slab
# into its own type") (in v5.17) moved slab information from struct page to
# struct slab. The former can be casted to the latter.
def _get_slab_type(prog: Program) -> Type:
    try:
        return prog.type("struct slab *")
    except LookupError:
        return prog.type("struct page *")


_OO_SHIFT = 16
_OO_MASK = (1 << _OO_SHIFT) - 1


def slab_cache_objects_per_slab(slab_cache: Object) -> int:
    """
    Get the number of objects in each slab of the given slab cache.

    This is only applicable to the SLUB and SLAB allocators; SLOB is not
    supported.

    :param slab_cache: ``struct kmem_cache *``
    """
    try:
        oo = slab_cache.oo
    except AttributeError:
        try:
            return slab_cache.num.value_()  # SLAB
        except AttributeError:
            raise ValueError("SLOB is not supported") from None
    else:
        return oo.x.value_() & _OO_MASK  # SLUB


def slab_cache_pages_per_slab(slab_cache: Object) -> int:
    """
    Get the number of pages allocated for each slab of the given slab cache.

    This is only applicable to the SLUB and SLAB allocators; SLOB is not
    supported.

    :param slab_cache: ``struct kmem_cache *``
    """
    return 1 << slab_cache_order(slab_cache)


def slab_cache_order(slab_cache: Object) -> int:
    """
    Get the allocation order (i.e., base 2 logarithm of the number of pages)
    for each slab of the given slab cache.

    This is only applicable to the SLUB and SLAB allocators; SLOB is not
    supported.

    >>> 1 << slab_cache_order(slab_cache) == slab_cache_pages_per_slab(slab_cache)
    True

    :param slab_cache: ``struct kmem_cache *``
    """
    try:
        oo = slab_cache.oo
    except AttributeError:
        try:
            return slab_cache.gfporder.value_()  # SLAB
        except AttributeError:
            raise ValueError("SLOB is not supported") from None
    else:
        return oo.x.value_() >> _OO_SHIFT  # SLUB


def slab_cache_is_merged(slab_cache: Object) -> bool:
    """
    Return whether a slab cache has been merged with any other slab caches.

    Unless configured otherwise, the kernel may merge slab caches of similar
    sizes together. See the `SLUB users guide
    <https://docs.kernel.org/vm/slub.html#slab-merging>`_ and
    ``slab_merge``/``slab_nomerge`` in the `kernel parameters documentation
    <https://www.kernel.org/doc/Documentation/admin-guide/kernel-parameters.txt>`_.

    This can cause confusion, as only the name of the first cache will be
    found, and objects of different types will be mixed in the same slab cache.

    For example, suppose that we have two types, ``struct foo`` and ``struct
    bar``, which have the same size but are otherwise unrelated. If the kernel
    creates a slab cache named ``foo`` for ``struct foo``, then another slab
    cache named ``bar`` for ``struct bar``, then slab cache ``foo`` will be
    reused instead of creating another cache for ``bar``. So the following will
    fail::

        find_slab_cache("bar")

    And the following will also return ``struct bar *`` objects errantly casted to
    ``struct foo *``::

        slab_cache_for_each_allocated_object(find_slab_cache("foo"), "struct foo")

    Unfortunately, these issues are difficult to work around generally, so one
    must be prepared to handle them on a case-by-case basis (e.g., by looking
    up the slab cache by its variable name and by checking that members of the
    structure make sense for the expected type).

    :param slab_cache: ``struct kmem_cache *``
    """
    return slab_cache.refcount > 1


@takes_program_or_default
def get_slab_cache_aliases(prog: Program) -> Dict[str, str]:
    """
    Return a dict mapping slab cache name to the cache it was merged with.

    The SLAB and SLUB subsystems can merge caches with similar settings and
    object sizes, as described in the documentation of
    :func:`slab_cache_is_merged()`. In some cases, the information about which
    caches were merged is lost, but in other cases, we can reconstruct the info.
    This function reconstructs the mapping, but requires that the kernel is
    configured with ``CONFIG_SLUB`` and ``CONFIG_SYSFS``.

    The returned dict maps from original cache name, to merged cache name. You
    can use this mapping to discover the correct cache to lookup via
    :func:`find_slab_cache()`. The dict contains an entry only for caches which
    were merged into a cache of a different name.

    >>> cache_to_merged = get_slab_cache_aliases()
    >>> cache_to_merged["dnotify_struct"]
    'avc_xperms_data'
    >>> "avc_xperms_data" in cache_to_merged
    False
    >>> find_slab_cache("dnotify_struct") is None
    True
    >>> find_slab_cache("avc_xperms_data") is None
    False

    :warning: This function will only work on kernels which are built with
      ``CONFIG_SLUB`` and ``CONFIG_SYSFS`` enabled.

    :returns: Mapping of slab cache name to final merged name
    :raises LookupError: If the helper fails because the debugged kernel
      doesn't have the required configuration
    """
    try:
        slab_kset = prog["slab_kset"]
    except KeyError:
        raise LookupError(
            "Couldn't find SLUB sysfs information: get_slab_cache_aliases() "
            "requires CONFIG_SLUB and CONFIG_SYSFS enabled in the debugged "
            "kernel."
        ) from None
    link_flag = prog.constant("KERNFS_LINK")
    name_map = {}
    for child in rbtree_inorder_for_each_entry(
        "struct kernfs_node",
        slab_kset.kobj.sd.dir.children.address_of_(),
        "rb",
    ):
        if child.flags & link_flag:
            cache = container_of(
                cast("struct kobject *", child.symlink.target_kn.priv),
                "struct kmem_cache",
                "kobj",
            )
            original_name = os.fsdecode(child.name.string_())
            target_name = os.fsdecode(cache.name.string_())
            if original_name != target_name:
                name_map[original_name] = target_name
    return name_map


@takes_program_or_default
def for_each_slab_cache(prog: Program) -> Iterator[Object]:
    """
    Iterate over all slab caches.

    :return: Iterator of ``struct kmem_cache *`` objects.
    """
    return list_for_each_entry(
        "struct kmem_cache", prog["slab_caches"].address_of_(), "list"
    )


@takes_program_or_default
def find_slab_cache(prog: Program, name: Union[str, bytes]) -> Optional[Object]:
    """
    Return the slab cache with the given name.

    :param name: Slab cache name.
    :return: ``struct kmem_cache *``
    """
    if isinstance(name, str):
        name = name.encode()
    for s in for_each_slab_cache(prog):
        if s.name.string_() == name:
            return s
    return None


@takes_program_or_default
def print_slab_caches(prog: Program) -> None:
    """Print the name and ``struct kmem_cache *`` value of all slab caches."""
    for s in for_each_slab_cache(prog):
        name = escape_ascii_string(s.name.string_(), escape_backslash=True)
        print(f"{name} ({s.type_.type_name()})0x{s.value_():x}")


class SlabCorruptionError(ValidationError):
    """
    Error raised when a corruption is encountered in a slab allocator data
    structure.
    """


class SlabFreelistCycleError(SlabCorruptionError):
    """
    Error raised when a cycle is encountered in a slab allocator freelist.
    """


class SlabPartialListError(SlabCorruptionError):
    """
    Error raised when a corruption is encountered in a list of partial slabs.
    """


# Get the name of a slab cache or fall back to a placeholder.
def _slab_cache_name(slab_cache: Object) -> str:
    try:
        return os.fsdecode(slab_cache.name.string_())
    except FaultError:
        return "slab cache"


# Between SLUB, SLAB, their respective configuration options, and the
# differences between kernel versions, there is a lot of state that we need to
# keep track of to inspect the slab allocator. It isn't pretty, but this class
# and its subclasses track all of that complexity so that we can share code
# between slab helpers.
class _SlabCacheHelper:
    def __init__(self, slab_cache: Object) -> None:
        self._prog = slab_cache.prog_
        self._slab_cache = slab_cache.read_()

    # Not all of the methods need this, so cache it lazily.
    @functools.cached_property
    def _slab_cache_size(self) -> int:
        return self._slab_cache.size.value_()

    def usage(self) -> "SlabCacheUsage":
        raise NotImplementedError()

    def _page_objects(
        self, page: Object, slab: Object, pointer_type: Type
    ) -> Iterator[Object]:
        raise NotImplementedError()

    def for_each_allocated_object(self, type: Union[str, Type]) -> Iterator[Object]:
        pointer_type = self._prog.pointer_type(self._prog.type(type))
        slab_type = _get_slab_type(self._prog)
        # Get the underlying implementation directly to avoid overhead on each
        # page.
        PageSlab = _get_PageSlab_impl(self._prog)
        for page in for_each_page(self._prog):
            try:
                if not PageSlab(page):
                    continue
            except FaultError:
                continue
            slab = cast(slab_type, page)
            if slab.slab_cache == self._slab_cache:
                yield from self._page_objects(page, slab, pointer_type)

    def object_info(self, page: Object, slab: Object, addr: int) -> "SlabObjectInfo":
        raise NotImplementedError()


class _SlubPerCpuInfo(NamedTuple):
    free_objects: Set[int]
    num_free_objects: int
    error: Optional[Exception]


class _SlabCacheHelperSlub(_SlabCacheHelper):
    @functools.cached_property
    def _red_left_pad(self) -> int:
        try:
            return self._slab_cache.red_left_pad.value_()
        except AttributeError:
            return 0

    def _slub_get_freelist(
        self, freelist_name: Callable[[], str], freelist: Object
    ) -> Set[int]:
        # In SLUB, the freelist is a linked list with the next pointer located
        # at ptr + slab_cache->offset.
        freelist_set: Set[int] = set()
        # This is racy. On live kernels, we retry a limited number of times.
        num_attempts = 1000 if (self._prog.flags & ProgramFlags.IS_LIVE) else 1
        for attempts_remaining in range(num_attempts, -1, -1):
            ptr = freelist.value_()
            while ptr:
                if ptr in freelist_set:
                    if attempts_remaining > 0:
                        # Break the loop over the freelist and retry from the
                        # beginning of the list.
                        break
                    e = SlabFreelistCycleError(
                        f"{_slab_cache_name(self._slab_cache)} {freelist_name()} "
                        "freelist contains cycle; "
                        "may be corrupted or in the middle of update"
                    )
                    # Smuggle the freelist we got so far.
                    e.freelist = freelist_set  # type: ignore[attr-defined]
                    raise e
                freelist_set.add(ptr)
                try:
                    ptr = self._freelist_dereference(ptr + self._freelist_offset)
                except FaultError as e:
                    if attempts_remaining > 0:
                        break
                    e.freelist = freelist_set  # type: ignore[attr-defined]
                    raise
            else:
                return freelist_set

            freelist_set.clear()
        assert False  # Tell mypy that this is unreachable.

    @functools.cached_property
    def _freelist_offset(self) -> int:
        return self._slab_cache.offset.value_()

    @functools.cached_property
    def _freelist_dereference(self) -> Callable[[int], int]:
        # If CONFIG_SLAB_FREELIST_HARDENED is enabled, then the next pointer is
        # obfuscated using slab_cache->random.
        try:
            freelist_random = self._slab_cache.random.value_()
        except AttributeError:
            return self._prog.read_word

        ulong_size = sizeof(self._prog.type("unsigned long"))

        # Since Linux kernel commit 1ad53d9fa3f6 ("slub: improve bit diffusion
        # for freelist ptr obfuscation") in v5.7, a swab() was added to the
        # freelist dereferencing calculation. This commit was backported to all
        # stable branches which have CONFIG_SLAB_FREELIST_HARDENED, but you can
        # still encounter some older stable kernels which don't have it.
        # Unfortunately, there's no easy way to detect whether it is in effect,
        # since the commit adds no struct field or other detectable difference.
        #
        # To handle this, we implement both methods, and we start out with a
        # "trial" function. On the first time we encounter a non-NULL freelist,
        # we try using the method with the swab(), and test whether the
        # resulting pointer may be dereferenced. If it can, we commit to using
        # that method forever. If it cannot, we switch to the version without
        # swab() and commit to using that.

        def _freelist_dereference_swab(ptr_addr: int) -> int:
            # *ptr_addr ^ slab_cache->random ^ byteswap(ptr_addr)
            return (
                self._prog.read_word(ptr_addr)
                ^ freelist_random
                ^ int.from_bytes(ptr_addr.to_bytes(ulong_size, "little"), "big")
            )

        def _freelist_dereference_no_swab(ptr_addr: int) -> int:
            # *ptr_addr ^ slab_cache->random ^ ptr_addr
            return self._prog.read_word(ptr_addr) ^ freelist_random ^ ptr_addr

        def _try_hardened_freelist_dereference(ptr_addr: int) -> int:
            result = _freelist_dereference_swab(ptr_addr)
            if result:
                try:
                    self._prog.read_word(result)
                    self._freelist_dereference = _freelist_dereference_swab
                except FaultError:
                    result = _freelist_dereference_no_swab(ptr_addr)
                    self._freelist_dereference = _freelist_dereference_no_swab
            return result

        return _try_hardened_freelist_dereference

    @staticmethod
    def _sheaf_get_objects(sheaf: Object, result: Set[int]) -> None:
        for object in sheaf.objects[: sheaf.size]:
            result.add(object.value_())

    def _sheaves_full_objects(self) -> Set[int]:
        result: Set[int] = set()
        for node in self._slab_cache.node[: nr_node_ids(self._prog)]:
            # struct kmem_cache_node::barn was added in Linux kernel commit
            # 2d517aa09bbc ("slab: add opt-in caching layer of percpu sheaves")
            # (in v6.18).
            try:
                barn = node.barn.read_()
            except AttributeError:
                return result
            if not barn:
                continue
            for sheaf in validate_list_for_each_entry(
                "struct slab_sheaf",
                barn.sheaves_full.address_of_(),
                "barn_list",
            ):
                self._sheaf_get_objects(sheaf, result)
        return result

    def _num_cpu_partial_list_free_objs(self, cpu: int, cpu_slab: Object) -> int:
        try:
            partial = cpu_slab.partial.read_()
        except AttributeError:
            # Since Linux kernel commit a93cf07bc3fb ("mm/slub.c: wrap
            # cpu_slab->partial in CONFIG_SLUB_CPU_PARTIAL") (in v4.13), if
            # CONFIG_SLUB_CPU_PARTIAL=n, then cpu_slab->partial does not exist.
            # Override the method on this instance to avoid checking again.
            self._num_cpu_partial_list_free_objs = lambda cpu, cpu_slab: 0  # type: ignore[method-assign]
            return 0

        # Since Linux kernel commit bb192ed9aa71 ("mm/slub: Convert most struct
        # page to struct slab by spatch") (in v5.17), the number of slabs on
        # the partial list is `->slabs`. Before that, it is `->pages`.
        nr_slabs_member = "slabs" if hasattr(partial, "slabs") else "pages"

        # This is racy. On live kernels, we retry a limited number of times.
        num_attempts = 1000 if (self._prog.flags & ProgramFlags.IS_LIVE) else 1
        for attempts_remaining in range(num_attempts, -1, -1):
            free_objs = 0
            prev_nr_slabs = None
            while partial:
                try:
                    nr_slabs = partial.member_(nr_slabs_member).value_()
                    # We could be stricter and check nr_slabs == prev_nr_slabs - 1, but
                    # the main thing we care about is not getting stuck in a cycle.
                    if prev_nr_slabs is not None and nr_slabs >= prev_nr_slabs:
                        if attempts_remaining > 0:
                            # Break the loop over the slab list and retry from the
                            # beginning of the list.
                            break
                        raise SlabPartialListError(
                            f"{_slab_cache_name(self._slab_cache)} cpu {cpu} "
                            "partial slabs count not decreasing; "
                            "may be corrupted or in the middle of update"
                        )
                    prev_nr_slabs = nr_slabs

                    free_objs += partial.objects.value_() - partial.inuse.value_()
                    partial = partial.next.read_()
                except FaultError:
                    if attempts_remaining > 0:
                        break
                    raise
            else:
                return free_objs

            partial = cpu_slab.partial.read_()
        assert False  # Tell mypy that this is unreachable.

    # Ugly method that gets all per-CPU information for our slab cache. If
    # count_free_objects is False, then this builds a set of free objects not
    # associated with a specific slab. If count_free_objects is True, then this
    # sums up the number of free objects in any per-CPU structures.
    def _per_cpu_info(
        self, *, count_free_objects: bool = False, catch_errors: bool = False
    ) -> _SlubPerCpuInfo:
        free_objects: Set[int] = set()
        num_free_objects = 0
        error = None

        # cpu_sheaves was added in Linux kernel commit 2d517aa09bbc ("slab: add
        # opt-in caching layer of percpu sheaves") (in v6.18).
        try:
            cpu_sheaves = self._slab_cache.cpu_sheaves.read_()
        except AttributeError:
            pass
        else:
            if cpu_sheaves:
                try:
                    for cpu in for_each_online_cpu(self._prog):
                        pcs = per_cpu_ptr(cpu_sheaves, cpu)
                        main = pcs.main.read_()
                        if main:
                            if count_free_objects:
                                num_free_objects += main.size.value_()
                            else:
                                self._sheaf_get_objects(main, free_objects)
                        spare = pcs.spare.read_()
                        if spare:
                            if count_free_objects:
                                num_free_objects += spare.size.value_()
                            else:
                                self._sheaf_get_objects(spare, free_objects)
                except (SlabCorruptionError, FaultError) as e:
                    if not catch_errors:
                        raise
                    error = e

        # cpu_slab was removed in Linux kernel commit 32c894c7274b ("slab:
        # remove struct kmem_cache_cpu") (in v7.0) in favor of cpu_sheaves.
        # Additionally, before Linux kernel commit 31e0886fd57d ("slub: remove
        # CONFIG_SLUB_TINY specific code paths") (in v6.19), cpu_slab didn't
        # exist on CONFIG_SLUB_TINY.
        try:
            cpu_slab = self._slab_cache.cpu_slab.read_()
        except AttributeError:
            pass
        else:
            # Since Linux kernel commit bb192ed9aa71 ("mm/slub: Convert most
            # struct page to struct slab by spatch") (in v5.17), the current
            # slab for a CPU is `struct slab *slab`. Before that, it is `struct
            # page *page`.
            cpu_slab_member = "slab" if hasattr(cpu_slab, "slab") else "page"
            try:
                for cpu in for_each_online_cpu(self._prog):
                    this_cpu_slab = per_cpu_ptr(cpu_slab, cpu)
                    slab = this_cpu_slab.member_(cpu_slab_member).read_()
                    if slab and slab.slab_cache == self._slab_cache:
                        freelist = self._slub_get_freelist(
                            lambda: f"cpu {cpu}", this_cpu_slab.freelist
                        )
                        if count_free_objects:
                            num_free_objects += len(freelist)
                        else:
                            free_objects |= freelist

                    if count_free_objects:
                        num_free_objects += self._num_cpu_partial_list_free_objs(
                            cpu, this_cpu_slab
                        )
            except (SlabCorruptionError, FaultError) as e:
                if not catch_errors:
                    raise
                error = e

        return _SlubPerCpuInfo(
            free_objects=free_objects,
            num_free_objects=num_free_objects,
            error=error,
        )

    def usage(self) -> "SlabCacheUsage":
        num_attempts = 1000 if (self._prog.flags & ProgramFlags.IS_LIVE) else 1

        num_slabs = 0
        num_objs = 0
        # SLUB doesn't maintain a counter of free objects. Instead, we have to
        # compute it from several places:
        #
        # 1. Per-node partial lists: sum of slab->objects - slab->inuse for
        #    each slab.
        # 2. Before Linux 7.0, per-CPU partial lists (ditto).
        # 3. Since Linux 6.18, per-node full sheaves: sheaf->size.
        # 4. Since Linux 6.18, per-CPU sheaves (ditto).
        # 5. Per-CPU freelists. We can't use slab->inuse of the per-CPU active
        #    slabs because it is set to slab->objects for active slabs.
        free_objs = self._per_cpu_info(count_free_objects=True).num_free_objects

        slab_type = _get_slab_type(self._prog).type
        # Since Linux kernel commit 4da1984edbbe ("mm: combine LRU and main
        # union in struct page") (in v4.18), the slab list_head is slab_list in
        # struct slab or struct page. Before that, it is struct page::lru.
        # (Between that commit and commit 916ac0527837 ("slub: use slab_list
        # instead of lru") (in v5.2), the SLUB code still uses struct page::lru,
        # but it is an alias of struct page::slab_list.)
        slab_list_member = "slab_list" if slab_type.has_member("slab_list") else "lru"
        for node in self._slab_cache.node[: nr_node_ids(self._prog)]:
            try:
                nr_slabs = node.nr_slabs
            except AttributeError:
                raise ValueError(
                    "slab_cache_usage() requires CONFIG_SLUB_DEBUG enabled in the kernel"
                ) from None
            num_slabs += nr_slabs.counter.value_()
            num_objs += node.total_objects.counter.value_()

            # struct kmem_cache_node::barn was added in Linux kernel commit
            # 2d517aa09bbc ("slab: add opt-in caching layer of percpu sheaves")
            # (in v6.18).
            try:
                barn = node.barn.read_()
            except AttributeError:
                pass
            else:
                if barn:
                    for sheaf in validate_list_for_each_entry(
                        "struct slab_sheaf",
                        barn.sheaves_full.address_of_(),
                        "barn_list",
                    ):
                        free_objs += sheaf.size.value_()

            # This is racy. On live kernels, we retry a limited number of times.
            for attempts_remaining in range(num_attempts, -1, -1):
                node_free_objs = 0
                try:
                    for slab in validate_list_for_each_entry(
                        slab_type, node.partial.address_of_(), slab_list_member
                    ):
                        node_free_objs += slab.objects.value_() - slab.inuse.value_()
                    break
                except (FaultError, ValidationError):
                    if attempts_remaining == 0:
                        raise
            free_objs += node_free_objs

        return SlabCacheUsage(
            num_slabs=num_slabs, num_objs=num_objs, free_objs=free_objs
        )

    def _page_objects(
        self, page: Object, slab: Object, pointer_type: Type
    ) -> Iterator[Object]:
        freelist = self._slub_get_freelist(lambda: f"slab {hex(slab)}", slab.freelist)
        addr = page_to_virt(page).value_() + self._red_left_pad
        end = addr + self._slab_cache_size * slab.objects
        while addr < end:
            if addr not in freelist and addr not in self._cache_free_objects:
                yield Object(self._prog, pointer_type, value=addr)
            addr += self._slab_cache_size

    def for_each_allocated_object(self, type: Union[str, Type]) -> Iterator[Object]:
        self._cache_free_objects = (
            self._sheaves_full_objects() | self._per_cpu_info().free_objects
        )
        return super().for_each_allocated_object(type)

    def object_info(self, page: Object, slab: Object, addr: int) -> "SlabObjectInfo":
        first_addr = page_to_virt(page).value_() + self._red_left_pad
        address = (
            first_addr
            + (addr - first_addr) // self._slab_cache_size * self._slab_cache_size
        )
        try:
            sheaves_full_objects = self._sheaves_full_objects()
        except (FaultError, ValidationError) as e:
            sheaves_full_objects = set()
            sheaves_error: Optional[Exception] = e
        else:
            sheaves_error = None
        per_cpu = self._per_cpu_info(catch_errors=True)
        if address in sheaves_full_objects or address in per_cpu.free_objects:
            allocated: Optional[bool] = False
        else:
            try:
                freelist = self._slub_get_freelist(
                    lambda: f"slab {hex(slab)}", slab.freelist
                )
            except (SlabCorruptionError, FaultError) as e:
                # On error, _slub_get_freelist() smuggles the partial freelist
                # it got as an attribute on the exception.
                allocated = False if address in getattr(e, "freelist", ()) else None
            else:
                if address in freelist:
                    allocated = False
                elif sheaves_error or per_cpu.error:
                    allocated = None
                else:
                    allocated = True
        return SlabObjectInfo(self._slab_cache, slab, address, allocated)


class _SlabCacheHelperSlab(_SlabCacheHelper):
    def __init__(self, slab_cache: Object) -> None:
        super().__init__(slab_cache)
        self._freelist_type = self._prog.type("freelist_idx_t *")

    @functools.cached_property
    def _obj_offset(self) -> int:
        try:
            return self._slab_cache.obj_offset.value_()
        except AttributeError:
            return 0

    @functools.cached_property
    def _slab_cache_num(self) -> int:
        return self._slab_cache.num.value_()

    @staticmethod
    def _slab_add_array_cache(array_caches: Set[int], ac: Object) -> None:
        for i in range(ac.avail):
            array_caches.add(ac.entry[i].value_())

    @functools.cached_property
    def _array_caches(self) -> Set[int]:
        array_caches: Set[int] = set()

        cpu_cache = self._slab_cache.cpu_cache.read_()
        for cpu in for_each_online_cpu(self._prog):
            self._slab_add_array_cache(array_caches, per_cpu_ptr(cpu_cache, cpu))

        nodes = self._slab_cache.node
        for nid in for_each_online_node(self._prog):
            n = nodes[nid].read_()
            shared = n.shared.read_()
            if shared:
                self._slab_add_array_cache(array_caches, shared)
            alien = n.alien.read_()
            if alien:
                for nid2 in for_each_online_node(self._prog):
                    self._slab_add_array_cache(array_caches, alien[nid2])

        return array_caches

    def usage(self) -> "SlabCacheUsage":
        num_slabs = 0
        # SLAB maintains per-node counters of free objects, but the array
        # caches are not included in those counters.
        free_objs = len(self._array_caches)

        for node in self._slab_cache.node[: nr_node_ids(self._prog)]:
            try:
                num_slabs += node.total_slabs.value_()
            except AttributeError:
                # Before Linux kernel commits f728b0a5d72a ("mm, slab: faster
                # active and free stats") and bf00bd345804 ("mm, slab: maintain
                # total slab count instead of active count") (in v4.10), struct
                # kmem_cache_node::total_slabs is named num_slabs instead.
                num_slabs += node.num_slabs.value_()
            free_objs += node.free_objects.value_()

        return SlabCacheUsage(
            num_slabs=num_slabs,
            num_objs=num_slabs * self._slab_cache_num,
            free_objs=free_objs,
        )

    def _slab_freelist(self, slab: Object) -> Set[int]:
        # In SLAB, the freelist is an array of free object indices.
        freelist = cast(self._freelist_type, slab.freelist)
        return {freelist[i].value_() for i in range(slab.active, self._slab_cache_num)}

    def _page_objects(
        self, page: Object, slab: Object, pointer_type: Type
    ) -> Iterator[Object]:
        freelist = self._slab_freelist(slab)
        s_mem = slab.s_mem.value_()
        for i in range(self._slab_cache_num):
            if i in freelist:
                continue
            addr = s_mem + i * self._slab_cache_size + self._obj_offset
            if addr in self._array_caches:
                continue
            yield Object(self._prog, pointer_type, value=addr)

    def object_info(self, page: Object, slab: Object, addr: int) -> "SlabObjectInfo":
        s_mem = slab.s_mem.value_()
        object_index = (addr - s_mem) // self._slab_cache_size
        object_address = s_mem + object_index * self._slab_cache_size
        return SlabObjectInfo(
            self._slab_cache,
            slab,
            object_address,
            allocated=object_address not in self._array_caches
            and object_index not in self._slab_freelist(slab),
        )


class _SlabCacheHelperSlob(_SlabCacheHelper):
    def usage(self) -> "SlabCacheUsage":
        raise ValueError("SLOB is not supported")

    def for_each_allocated_object(self, type: Union[str, Type]) -> Iterator[Object]:
        raise ValueError("SLOB is not supported")

    def object_info(self, page: Object, slab: Object, addr: int) -> "SlabObjectInfo":
        return SlabObjectInfo(self._slab_cache, slab, 0, None)


def _get_slab_cache_helper(slab_cache: Object) -> _SlabCacheHelper:
    prog = slab_cache.prog_
    try:
        type = prog.cache["slab_cache_helper_type"]
    except KeyError:
        try:
            prog.type("freelist_idx_t *")
            type = _SlabCacheHelperSlab
        except LookupError:
            if hasattr(slab_cache, "offset"):
                type = _SlabCacheHelperSlub
            else:
                type = _SlabCacheHelperSlob
        prog.cache["slab_cache_helper_type"] = type
    return type(slab_cache)


def slab_cache_usage(slab_cache: Object) -> "SlabCacheUsage":
    """
    Get statistics for how many slabs and objects are allocated for a given
    slab cache.

    Only the SLUB and SLAB allocators are supported; SLOB does not track these
    statistics. Additionally, for SLUB, ``CONFIG_SLUB_DEBUG`` must be enabled
    in the kernel (this is the default unless ``CONFIG_SLUB_TINY`` is enabled).

    :param slab_cache: ``struct kmem_cache *``
    """
    return _get_slab_cache_helper(slab_cache).usage()


class SlabCacheUsage(NamedTuple):
    """Slab cache usage statistics returned by :func:`slab_cache_usage()`."""

    num_slabs: int
    """Number of slabs allocated for this slab cache."""

    num_objs: int
    """Total number of objects in this slab cache (free and active)."""

    free_objs: int
    """Number of free objects in this slab cache."""

    @property
    def active_objs(self) -> int:
        """Number of active (allocated) objects in this slab cache."""
        return self.num_objs - self.free_objs


def slab_cache_for_each_allocated_object(
    slab_cache: Object, type: Union[str, Type]
) -> Iterator[Object]:
    """
    Iterate over all allocated objects in a given slab cache.

    Only the SLUB and SLAB allocators are supported; SLOB does not store enough
    information to identify objects in a slab cache.

    >>> dentry_cache = find_slab_cache("dentry")
    >>> next(slab_cache_for_each_allocated_object(dentry_cache, "struct dentry"))
    *(struct dentry *)0xffff905e41404000 = {
        ...
    }

    :param slab_cache: ``struct kmem_cache *``
    :param type: Type of object in the slab cache.
    :return: Iterator of ``type *`` objects.
    """
    return _get_slab_cache_helper(slab_cache).for_each_allocated_object(type)


def _find_containing_slab(prog: Program, addr: int) -> Optional[Tuple[Object, Object]]:
    page = virt_to_page(prog, addr)

    try:
        page = compound_head(page)
        if not PageSlab(page):
            return None
    except FaultError:
        # Page does not exist
        return None

    return page, cast(_get_slab_type(prog), page)


@takes_program_or_default
def slab_object_info(prog: Program, addr: IntegerLike) -> "Optional[SlabObjectInfo]":
    """
    Get information about an address if it is in a slab object.

    >>> ptr = find_task(1).comm.address_of_()
    >>> info = slab_object_info(ptr)
    >>> info
    SlabObjectInfo(slab_cache=Object(prog, 'struct kmem_cache *', address=0xffffdb93c0045e18), slab=Object(prog, 'struct slab *', value=0xffffdb93c0045e00), address=0xffffa2bf81178000, allocated=True)

    Note that :attr:`SlabObjectInfo.address` is the start address of the
    object, which may be less than *addr* if *addr* points to a member inside
    of the object:

    >>> ptr.value_() - info.address
    1496
    >>> offsetof(prog.type("struct task_struct"), "comm")
    1496

    Note that SLOB does not store enough information to identify slab objects,
    so if the kernel is configured to use SLOB, then
    :attr:`SlabObjectInfo.slab_cache` will always be ``NULL`` and
    :attr:`SlabObjectInfo.address` will always be 0. Additionally, for
    allocations of at least one page, SLOB allocates pages directly, so this
    will return ``None``.

    :param addr: ``void *``
    :return: :class:`SlabObjectInfo` if *addr* is in a slab object, or ``None``
        if not.
    """
    addr = operator.index(addr)
    if not in_direct_map(prog, addr):
        return None
    result = _find_containing_slab(prog, addr)
    if result is None:
        return None
    page, slab = result
    try:
        slab_cache = slab.slab_cache.read_()
    except AttributeError:
        # SLOB
        slab_cache = NULL(prog, "struct kmem_cache *")
    return _get_slab_cache_helper(slab_cache).object_info(page, slab, addr)


class SlabObjectInfo:
    """Information about an object in the slab allocator."""

    slab_cache: Object
    """
    ``struct kmem_cache *`` that the slab object is from.

    SLOB does not store enough information to find this, so if the kernel is
    configured to use SLOB, then this will always be ``NULL``.
    """

    slab: Object
    """
    Slab containing the slab object.

    Since Linux v5.17, this is a ``struct slab *``. Before that, it is a
    ``struct page *``.
    """

    address: int
    """
    Address of the slab object.

    SLOB does not store enough information to find this, so if the kernel is
    configured to use SLOB, then this will always be 0.
    """

    allocated: Optional[bool]
    """
    ``True`` if the object is allocated, ``False`` if it is free, or ``None``
    if not known because the slab cache is corrupted or the kernel is
    configured to use SLOB.
    """

    def __init__(
        self, slab_cache: Object, slab: Object, address: int, allocated: Optional[bool]
    ) -> None:
        self.slab_cache = slab_cache
        self.slab = slab
        self.address = address
        self.allocated = allocated

    def __repr__(self) -> str:
        return f"SlabObjectInfo(slab_cache={self.slab_cache!r}, slab={self.slab!r}, address={hex(self.address)}, allocated={self.allocated})"


@takes_program_or_default
def find_containing_slab_cache(prog: Program, addr: IntegerLike) -> Object:
    """
    Get the slab cache that an address was allocated from, if any.

    Note that SLOB does not store enough information to identify objects in a
    slab cache, so if the kernel is configured to use SLOB, this will always
    return ``NULL``.

    :param addr: ``void *``
    :return: ``struct kmem_cache *`` containing *addr*, or ``NULL`` if *addr*
        is not from a slab cache.
    """
    if not in_direct_map(prog, addr):
        return NULL(prog, "struct kmem_cache *")
    result = _find_containing_slab(prog, operator.index(addr))
    if result is not None:
        try:
            return result[1].slab_cache.read_()
        except AttributeError:
            # SLOB
            pass
    return NULL(prog, "struct kmem_cache *")


@takes_program_or_default
def slab_total_usage(prog: Program) -> "SlabTotalUsage":
    """Get the total number of reclaimable and unreclaimable slab pages."""
    # The items were renamed in Linux kernel commit d42f3245c7e2 ("mm: memcg:
    # convert vmstat slab counters to bytes") (in v5.9).
    try:
        return SlabTotalUsage(
            reclaimable_pages=global_node_page_state(prog["NR_SLAB_RECLAIMABLE_B"]),
            unreclaimable_pages=global_node_page_state(prog["NR_SLAB_UNRECLAIMABLE_B"]),
        )
    except ObjectNotFoundError:
        return SlabTotalUsage(
            reclaimable_pages=global_node_page_state(prog["NR_SLAB_RECLAIMABLE"]),
            unreclaimable_pages=global_node_page_state(prog["NR_SLAB_UNRECLAIMABLE"]),
        )


class SlabTotalUsage(NamedTuple):
    """Slab memory usage returned by :func:`slab_total_usage()`."""

    reclaimable_pages: int
    """Number of reclaimable slab pages."""

    unreclaimable_pages: int
    """Number of unreclaimable slab pages."""

    @property
    def total_pages(self) -> int:
        """Total number of slab pages."""
        return self.reclaimable_pages + self.unreclaimable_pages



# Copyright (c) Meta Platforms, Inc. and affiliates.
# Copyright (c) 2026, Oracle and/or its affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
CPU Scheduler
-------------

The ``drgn.helpers.linux.sched`` module provides helpers for working with the
Linux CPU scheduler.
"""

from typing import Iterator, Tuple

from _drgn import (
    _linux_helper_cpu_curr,
    _linux_helper_idle_task,
    _linux_helper_task_cpu as task_cpu,
    _linux_helper_task_on_cpu as task_on_cpu,
    _linux_helper_task_thread_info as task_thread_info,
)
from drgn import IntegerLike, Object, Program, container_of
from drgn.helpers.common.prog import takes_program_or_default
from drgn.helpers.linux.cgroup import cgroup_name
from drgn.helpers.linux.list import list_for_each_entry
from drgn.helpers.linux.percpu import per_cpu
from drgn.helpers.linux.rbtree import rbtree_inorder_for_each_entry

__all__ = (
    "cfs_rq_for_each_entity",
    "cpu_curr",
    "cpu_rq",
    "get_task_state",
    "idle_task",
    "loadavg",
    "rq_for_each_fair_task",
    "rq_for_each_rt_task",
    "sched_entity_is_task",
    "sched_entity_to_task",
    "task_cpu",
    "task_group_name",
    "task_on_cpu",
    "task_rq",
    "task_since_last_arrival_ns",
    "task_state_to_char",
    "task_thread_info",
    "thread_group_leader",
)

_TASK_NOLOAD = 0x400


@takes_program_or_default
def cpu_curr(prog: Program, cpu: IntegerLike) -> Object:
    """
    Return the task running on the given CPU.

    >>> cpu_curr(7).comm
    (char [16])"python3"

    :param cpu: CPU number.
    :return: ``struct task_struct *``
    """
    return _linux_helper_cpu_curr(prog, cpu)


@takes_program_or_default
def idle_task(prog: Program, cpu: IntegerLike) -> Object:
    """
    Return the idle thread (PID 0, a.k.a swapper) for the given CPU.

    >>> idle_task(1).comm
    (char [16])"swapper/1"

    :param cpu: CPU number.
    :return: ``struct task_struct *``
    """
    return _linux_helper_idle_task(prog, cpu)


def task_state_to_char(task: Object) -> str:
    """
    Get the state of the task as a character (e.g., ``'R'`` for running). See
    `ps(1)
    <http://man7.org/linux/man-pages/man1/ps.1.html#PROCESS_STATE_CODES>`_ for
    a description of the process state codes.

    :param task: ``struct task_struct *``
    """
    prog = task.prog_
    task_state_chars: str
    TASK_REPORT: int
    try:
        task_state_chars, TASK_REPORT, task_state_name = prog.cache[
            "task_state_to_char"
        ]
    except KeyError:
        task_state_array = prog["task_state_array"]
        # Walk through task_state_array backwards looking for the largest state
        # that we know is in TASK_REPORT, then populate the task state mapping.
        chars = None
        for i in range(len(task_state_array) - 1, -1, -1):
            c: int = task_state_array[i][0].value_()
            if chars is None and c in b"RSDTtXZP":
                chars = bytearray(i + 1)
                TASK_REPORT = (1 << i) - 1
            if chars is not None:
                chars[i] = c
        if chars is None:
            raise Exception("could not parse task_state_array")
        task_state_chars = chars.decode("ascii")

        # Since Linux kernel commit 2f064a59a11f ("sched: Change
        # task_struct::state") (in v5.14), the task state is named "__state".
        # Before that, it is named "state".
        try:
            task_state = task.__state
            task_state_name = "__state"
        except AttributeError:
            task_state = task.state
            task_state_name = "state"

        prog.cache["task_state_to_char"] = (
            task_state_chars,
            TASK_REPORT,
            task_state_name,
        )
    else:
        task_state = task.member_(task_state_name)
    task_state = task_state.value_()
    exit_state = task.exit_state.value_()
    state = (task_state | exit_state) & TASK_REPORT
    char = task_state_chars[state.bit_length()]
    # States beyond TASK_REPORT are special. As of Linux v5.14, TASK_IDLE is
    # the only one; it is defined as TASK_UNINTERRUPTIBLE | TASK_NOLOAD.
    if char == "D" and (task_state & ~state) == _TASK_NOLOAD:
        return "I"
    else:
        return char


_TASK_STATE_CHAR_TO_STATE = {
    "R": "R (running)",
    "S": "S (sleeping)",
    "D": "D (disk sleep)",
    "T": "T (stopped)",
    "t": "t (tracing stop)",
    "X": "X (dead)",
    "Z": "Z (zombie)",
    "P": "P (parked)",
    "I": "I (idle)",
}


def get_task_state(task: Object) -> str:
    """
    Get the state of the task as a character plus a parenthesized name (e.g.,
    ``'R (running)'``).

    See also :func:`task_state_to_char()`.

    :param task: ``struct task_struct *``
    """
    char = task_state_to_char(task)
    return _TASK_STATE_CHAR_TO_STATE.get(char, char)


@takes_program_or_default
def loadavg(prog: Program) -> Tuple[float, float, float]:
    """
    Return system load averaged over 1, 5 and 15 minutes as
    tuple of three float values.

    >>> loadavg()
    (2.34, 0.442, 1.33)
    """

    avenrun = prog["avenrun"]
    vals = [avenrun[i].value_() / (1 << 11) for i in range(3)]
    return (vals[0], vals[1], vals[2])


@takes_program_or_default
def cpu_rq(prog: Program, cpu: IntegerLike) -> Object:
    """
    Get the runqueue for a given cpu.

    :param cpu: CPU number.
    :returns: ``struct rq *``
    """
    return per_cpu(prog["runqueues"], cpu).address_of_()


def task_rq(task: Object) -> Object:
    """
    Get the runqueue for a given task.

    :param task: ``struct task_struct *``
    :returns: ``struct rq *``
    """
    return cpu_rq(task.prog_, task_cpu(task))


def rq_for_each_fair_task(rq: Object) -> Iterator[Object]:
    """
    Iterate over tasks on a runqueue in the fair scheduling class (EEVDF or
    CFS).

    :param rq: ``struct rq *``
    :return: Iterator of ``struct task_struct *`` objects
    """
    try:
        cfs_tasks = rq.cfs_tasks
    except AttributeError:
        # Before Linux kernel commit cac5cefbade9 ("sched/smp: Make SMP
        # unconditional") (in v6.17), cfs_tasks does not exist on !SMP. Fall
        # back to walking the tasks_timeline hierarchy.
        return (
            sched_entity_to_task(se)
            for se, _, _, is_task in cfs_rq_for_each_entity(rq.cfs.address_of_())
            if is_task
        )
    return list_for_each_entry(
        "struct task_struct", cfs_tasks.address_of_(), "se.group_node"
    )


def rq_for_each_rt_task(rq: Object) -> Iterator[Object]:
    """
    Iterate over tasks on a runqueue in the realtime scheduling class.

    :param rq: ``struct rq *``
    :return: Iterator of ``struct task_struct *`` objects
    """
    for queue in rq.rt.active.queue:
        yield from list_for_each_entry(
            "struct task_struct", queue.address_of_(), "rt.run_list"
        )


def sched_entity_is_task(se: Object) -> bool:
    """
    Return whether a scheduler entity is a task.

    :param se: ``struct sched_entity *``
    """
    try:
        return not se.my_q
    except AttributeError:
        return True


def sched_entity_to_task(se: Object) -> Object:
    """
    Get the task represented by a scheduler entity.

    Note that this does not check whether the entity is actually a task; see
    :func:`sched_entity_is_task()`.

    :param se: ``struct sched_entity *``
    :return: ``struct task_struct *``
    """
    return container_of(se, "struct task_struct", "se")


def task_group_name(tg: Object) -> bytes:
    """
    Get the name of a task group.

    :param tg: ``struct task_group *``
    :return: Cgroup name, or empty byte string if not a cgroup.
    """
    cgrp = tg.css.cgroup.read_()
    if not cgrp:
        return b""
    return cgroup_name(cgrp)


def cfs_rq_for_each_entity(cfs_rq: Object) -> Iterator[Tuple[Object, int, bool, bool]]:
    """
    Iterate over all entities on a fair scheduler runqueue (recursively).

    Task groups are visited before their children (i.e., this does a pre-order
    traversal).

    :param cfs_rq: ``struct cfs_rq *``
    :return: Iterator of (``struct sched_entity *``, ``depth``, ``is_curr``,
        ``is_task``) tuples.

        ``depth`` is the depth of the entity relative to the runqueue (e.g.,
        direct children have depth 0, entities in descendant task groups have
        depth > 0).

        ``is_curr`` is whether the entity is the current entity on its
        runqueue.

        If ``is_task`` is ``True``, then the entity is a task, which can be
        obtained with :func:`sched_entity_to_task()`. Otherwise, it is a task
        group.
    """

    def entities(cfs_rq: Object) -> Iterator[Tuple[Object, bool]]:
        curr = cfs_rq.curr.read_()
        if curr:
            yield curr, True

        # Since Linux kernel commit bfb068892d30 ("sched/fair: replace
        # cfs_rq->rb_leftmost") (in v4.14), tasks_timeline is a struct
        # rb_root_cached. Before that, it was a struct rb_root.
        rb_root = cfs_rq.tasks_timeline
        try:
            rb_root = rb_root.rb_root
        except AttributeError:
            pass

        for se in rbtree_inorder_for_each_entry(
            "struct sched_entity", rb_root.address_of_(), "run_node"
        ):
            yield se, False

    stack = [(entities(cfs_rq), 0)]
    while stack:
        it, depth = stack[-1]
        try:
            se, is_curr = next(it)
        except StopIteration:
            stack.pop()
        else:
            try:
                my_q = se.my_q.read_()
            except AttributeError:
                my_q = None

            yield se, depth, is_curr, not my_q

            if my_q:
                stack.append((entities(my_q), depth + 1))


def task_since_last_arrival_ns(task: Object) -> int:
    """
    Get the difference between the runqueue timestamp when a task last started
    running and the current runqueue timestamp.

    This is approximately the time that the task has been in its current status
    (running, queued, or blocked). However, if a CPU is either idle or running
    the same task for a long time, then the timestamps will not be accurate.

    This is only supported if the kernel was compiled with
    ``CONFIG_SCHEDSTATS`` or ``CONFIG_TASK_DELAY_ACCT``.

    :param task: ``struct task_struct *``
    :returns: Duration in nanoseconds.
    """
    arrival_time = task.sched_info.last_arrival.value_()
    rq_clock = task_rq(task).clock.value_()

    return rq_clock - arrival_time


def thread_group_leader(task: Object) -> bool:
    """
    Return whether a task is a thread group leader.

    :param task: ``struct task_struct *``
    """
    return task.exit_signal >= 0



#!/usr/bin/env drgn
# Copyright (c) SUSE Linux.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""Print memory map of a given task."""

import os
import sys

from drgn.helpers.linux.device import MAJOR, MINOR
from drgn.helpers.linux.fs import d_path
from drgn.helpers.linux.mm import for_each_vma
from drgn.helpers.linux.pid import find_task

if len(sys.argv) != 2:
    sys.exit("Usage: ./vmmap.py PID")
pid = int(sys.argv[1])

task = find_task(prog, int(pid))
if not task:
    sys.exit(f"Cannot find task {pid}")

FLAGS = ((0x1, "r"), (0x2, "w"), (0x4, "x"))
PAGE_SHIFT = prog["PAGE_SHIFT"]

print("Start        End          Flgs   Offset Dev   Inode            File path")

for vma in for_each_vma(task.mm):
    flags = "".join([v if f & vma.vm_flags else "-" for f, v in FLAGS])
    flags += "s" if vma.vm_flags & 0x8 else "p"
    print(f"{vma.vm_start.value_():0x}-{vma.vm_end.value_():0x} {flags} ",
          end="")

    vmfile = vma.vm_file
    if vmfile:
        inode = vmfile.f_inode.i_ino.value_()
        dev = vmfile.f_inode.i_sb.s_dev
        major, minor = MAJOR(dev), MINOR(dev)
        path = os.fsdecode(d_path(vmfile.f_path))
        pgoff = (vma.vm_pgoff << PAGE_SHIFT).value_()
    else:
        inode = 0
        major, minor = 0, 0
        path = ""
        pgoff = 0

    print(f"{pgoff:08x} {major:02x}:{minor:02x} {inode:<16} {path}")



# Copyright (c) Meta Platforms, Inc. and affiliates.
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Process IDS
-----------

The ``drgn.helpers.linux.pid`` module provides helpers for looking up process
IDs and processes.
"""

from typing import Iterator, Optional

from _drgn import (
    _linux_helper_find_pid,
    _linux_helper_find_task,
    _linux_helper_pid_task as pid_task,
)
from drgn import IntegerLike, Object, Program, cast, container_of
from drgn.helpers.common.prog import takes_object_or_program_or_default
from drgn.helpers.linux.cpumask import for_each_online_cpu
from drgn.helpers.linux.idr import idr_for_each
from drgn.helpers.linux.list import hlist_for_each_entry, list_for_each_entry
from drgn.helpers.linux.sched import idle_task

__all__ = (
    "find_pid",
    "find_task",
    "for_each_pid",
    "for_each_task",
    "for_each_task_in_group",
    "pid_task",
)


@takes_object_or_program_or_default
def find_pid(prog: Program, ns: Optional[Object], pid: IntegerLike) -> Object:
    """
    Return the ``struct pid *`` for the given PID number.

    :param ns: ``struct pid_namespace *``. Defaults to the initial PID
        namespace if given a :class:`~drgn.Program` or :ref:`omitted
        <default-program>`.
    :return: ``struct pid *``
    """
    if ns is None:
        ns = prog["init_pid_ns"].address_of_()
    return _linux_helper_find_pid(ns, pid)


@takes_object_or_program_or_default
def for_each_pid(prog: Program, ns: Optional[Object]) -> Iterator[Object]:
    """
    Iterate over all PIDs in a namespace.

    :param ns: ``struct pid_namespace *``. Defaults to the initial PID
        namespace if given a :class:`~drgn.Program` or :ref:`omitted
        <default-program>`.
    :return: Iterator of ``struct pid *`` objects.
    """
    if ns is None:
        ns = prog["init_pid_ns"].address_of_()
    if hasattr(ns, "idr"):
        for nr, entry in idr_for_each(ns.idr):
            yield cast("struct pid *", entry)
    else:
        pid_hash = prog["pid_hash"]
        for i in range(1 << prog["pidhash_shift"].value_()):
            for upid in hlist_for_each_entry(
                "struct upid", pid_hash[i].address_of_(), "pid_chain"
            ):
                if upid.ns == ns:
                    yield container_of(upid, "struct pid", f"numbers[{int(ns.level)}]")


@takes_object_or_program_or_default
def find_task(prog: Program, ns: Optional[Object], pid: IntegerLike) -> Object:
    """
    Return the task with the given PID.

    :param ns: ``struct pid_namespace *``. Defaults to the initial PID
        namespace if given a :class:`~drgn.Program` or :ref:`omitted
        <default-program>`.
    :return: ``struct task_struct *``
    """
    if ns is None:
        ns = prog["init_pid_ns"].address_of_()
    return _linux_helper_find_task(ns, pid)


@takes_object_or_program_or_default
def for_each_task(
    prog: Program, ns: Optional[Object], *, idle: bool = False
) -> Iterator[Object]:
    """
    Iterate over all of the tasks visible in a namespace.

    :param ns: ``struct pid_namespace *``. Defaults to the initial PID
        namespace if given a :class:`~drgn.Program` or :ref:`omitted
        <default-program>`.
    :param idle: Whether to include the idle threads (PID 0, a.k.a swapper) for
        each CPU.
    :return: Iterator of ``struct task_struct *`` objects.
    """
    if idle:
        for cpu in for_each_online_cpu(prog):
            yield idle_task(prog, cpu)

    PIDTYPE_PID = prog["PIDTYPE_PID"].value_()
    for pid in for_each_pid(prog if ns is None else ns):
        task = pid_task(pid, PIDTYPE_PID)
        if task:
            yield task


def for_each_task_in_group(
    task: Object, include_self: bool = False
) -> Iterator[Object]:
    """
    Iterate over all tasks in the thread group

    Or, in the more common userspace terms, iterate over all threads of a
    process.

    :param task: a task whose group to iterate over
    :param include_self: should ``task`` itself be returned?
    :returns: an iterable of every thread in the thread group
    """
    for other in list_for_each_entry(
        "struct task_struct",
        task.signal.thread_head.address_of_(),
        "thread_node",
    ):
        if other != task or include_self:
            yield other
